"""Open index action class"""

import logging
from curator.helpers.testers import verify_index_list
from curator.helpers.utils import chunk_index_list, report_failure, show_dry_run, to_csv


class Open:
    """Open Action Class"""

    def __init__(self, ilo):
        """
        :param ilo: An IndexList Object

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        """
        verify_index_list(ilo)
        #: The :py:class:`~.curator.indexlist.IndexList` object passed from
        #: param ``ilo``
        self.index_list = ilo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        self.loggit = logging.getLogger('curator.actions.open')

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        show_dry_run(self.index_list, 'open')

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.IndicesClient.open` indices in
        :py:attr:`index_list`
        """
        self.index_list.empty_list_check()
        msg = (
            f'Opening {len(self.index_list.indices)} selected indices: '
            f'{self.index_list.indices}'
        )
        self.loggit.info(msg)
        try:
            index_lists = chunk_index_list(self.index_list.indices)
            for lst in index_lists:
                self.client.indices.open(index=to_csv(lst))
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)



"""Open (closed) Index Singleton"""

import click
from curator.cli_singletons.object_class import CLIAction
from curator.cli_singletons.utils import validate_filter_json


@click.command(name='open')
@click.option(
    '--search_pattern', type=str, default='*', help='Elasticsearch Index Search Pattern'
)
@click.option(
    '--ignore_empty_list',
    is_flag=True,
    help='Do not raise exception if there are no actionable indices',
)
@click.option(
    '--allow_ilm_indices/--no-allow_ilm_indices',
    help='Allow Curator to operate on Index Lifecycle Management monitored indices.',
    default=False,
    show_default=True,
)
@click.option(
    '--include_hidden/--no-include_hidden',
    help='Allow Curator to operate on hidden indices (and data_streams).',
    default=False,
    show_default=True,
)
@click.option(
    '--filter_list',
    callback=validate_filter_json,
    help='JSON array of filters selecting indices to act on.',
    required=True,
)
@click.pass_context
def open_indices(
    ctx,
    search_pattern,
    ignore_empty_list,
    allow_ilm_indices,
    include_hidden,
    filter_list,
):
    """
    Open Indices
    """
    # ctx.info_name is the name of the function or name specified in
    # @click.command decorator
    action = CLIAction(
        ctx.info_name,
        ctx.obj['configdict'],
        {
            'search_pattern': search_pattern,
            'allow_ilm_indices': allow_ilm_indices,
            'include_hidden': include_hidden,
        },
        filter_list,
        ignore_empty_list,
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])



"""Set up voluptuous Schema defaults for various actions"""

from voluptuous import Schema
from curator.defaults import option_defaults


# Methods for building the schema
def action_specific(action):
    """
    :param action: The name of an action
    :type action: str

    :returns: A :py:class:`list` containing one or more
        :py:class:`~.voluptuous.schema_builder.Optional` or
        :py:class:`~.voluptuous.schema_builder.Required` options from
        :py:mod:`~.curator.defaults.option_defaults`, defining acceptable values
        for each for the given ``action``
    :rtype: list
    """
    options = {
        'alias': [
            option_defaults.name(action),
            option_defaults.warn_if_no_indices(),
            option_defaults.extra_settings(),
        ],
        'allocation': [
            option_defaults.search_pattern(),
            option_defaults.key(),
            option_defaults.value(),
            option_defaults.allocation_type(),
            option_defaults.wait_for_completion(action),
            option_defaults.wait_interval(action),
            option_defaults.max_wait(action),
        ],
        'close': [
            option_defaults.search_pattern(),
            option_defaults.delete_aliases(),
            option_defaults.skip_flush(),
        ],
        'cluster_routing': [
            option_defaults.routing_type(),
            option_defaults.cluster_routing_setting(),
            option_defaults.cluster_routing_value(),
            option_defaults.wait_for_completion(action),
            option_defaults.wait_interval(action),
            option_defaults.max_wait(action),
        ],
        'cold2frozen': [
            option_defaults.search_pattern(),
            option_defaults.c2f_index_settings(),
            option_defaults.c2f_ignore_index_settings(),
            option_defaults.wait_for_completion('cold2frozen'),
        ],
        'create_index': [
            option_defaults.name(action),
            option_defaults.ignore_existing(),
            option_defaults.extra_settings(),
        ],
        'delete_indices': [
            option_defaults.search_pattern(),
        ],
        'delete_snapshots': [
            option_defaults.repository(),
            option_defaults.retry_interval(),
            option_defaults.retry_count(),
        ],
        'forcemerge': [
            option_defaults.search_pattern(),
            option_defaults.delay(),
            option_defaults.max_num_segments(),
        ],
        'index_settings': [
            option_defaults.search_pattern(),
            option_defaults.index_settings(),
            option_defaults.ignore_unavailable(),
            option_defaults.preserve_existing(),
        ],
        'open': [
            option_defaults.search_pattern(),
        ],
        'reindex': [
            option_defaults.request_body(),
            option_defaults.refresh(),
            option_defaults.requests_per_second(),
            option_defaults.slices(),
            option_defaults.timeout(action),
            option_defaults.wait_for_active_shards(action),
            option_defaults.wait_for_completion(action),
            option_defaults.wait_interval(action),
            option_defaults.max_wait(action),
            option_defaults.remote_certificate(),
            option_defaults.remote_client_cert(),
            option_defaults.remote_client_key(),
            option_defaults.remote_filters(),
            option_defaults.migration_prefix(),
            option_defaults.migration_suffix(),
        ],
        'replicas': [
            option_defaults.search_pattern(),
            option_defaults.count(),
            option_defaults.wait_for_completion(action),
            option_defaults.wait_interval(action),
            option_defaults.max_wait(action),
        ],
        'rollover': [
            option_defaults.name(action),
            option_defaults.new_index(),
            option_defaults.conditions(),
            option_defaults.extra_settings(),
            option_defaults.wait_for_active_shards(action),
        ],
        'restore': [
            option_defaults.repository(),
            option_defaults.name(action),
            option_defaults.indices(),
            option_defaults.ignore_unavailable(),
            option_defaults.include_aliases(),
            option_defaults.include_global_state(action),
            option_defaults.partial(),
            option_defaults.rename_pattern(),
            option_defaults.rename_replacement(),
            option_defaults.extra_settings(),
            option_defaults.wait_for_completion(action),
            option_defaults.wait_interval(action),
            option_defaults.max_wait(action),
            option_defaults.skip_repo_fs_check(),
        ],
        'snapshot': [
            option_defaults.search_pattern(),
            option_defaults.repository(),
            option_defaults.name(action),
            option_defaults.ignore_unavailable(),
            option_defaults.include_global_state(action),
            option_defaults.partial(),
            option_defaults.wait_for_completion(action),
            option_defaults.wait_interval(action),
            option_defaults.max_wait(action),
            option_defaults.skip_repo_fs_check(),
        ],
        'shrink': [
            option_defaults.search_pattern(),
            option_defaults.shrink_node(),
            option_defaults.node_filters(),
            option_defaults.number_of_shards(),
            option_defaults.number_of_replicas(),
            option_defaults.shrink_prefix(),
            option_defaults.shrink_suffix(),
            option_defaults.copy_aliases(),
            option_defaults.delete_after(),
            option_defaults.post_allocation(),
            option_defaults.wait_for_active_shards(action),
            option_defaults.extra_settings(),
            option_defaults.wait_for_completion(action),
            option_defaults.wait_for_rebalance(),
            option_defaults.wait_interval(action),
            option_defaults.max_wait(action),
        ],
    }
    return options[action]


def get_schema(action):
    """
    Return a :py:class:`~.voluptuous.schema_builder.Schema` of acceptable options
    and their default values as returned by :py:func:`action_specific`, passing
    along the value of ``action``.

    :param action: The name of an action
    :type action: str

    :returns: A valid :py:class:`~.voluptuous.schema_builder.Schema` of the options
        for ``action``
    """
    options = {}
    defaults = [
        option_defaults.allow_ilm_indices(),
        option_defaults.continue_if_exception(),
        option_defaults.disable_action(),
        option_defaults.ignore_empty_list(),
        option_defaults.include_hidden(),
        option_defaults.timeout_override(action),
    ]
    for each in defaults:
        options.update(each)
    for each in action_specific(action):
        options.update(each)
    return Schema(options)



"""Test snapshot action functionality"""

# pylint: disable=missing-function-docstring, missing-class-docstring, line-too-long
import os
from datetime import datetime, timedelta, timezone
from curator.helpers.getters import get_indices, get_snapshot
from curator.exceptions import FailedExecution
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


class TestActionFileSnapshot(CuratorTestCase):
    def test_snapshot(self):
        self.create_indices(5)
        self.create_repository()
        snap_name = 'snapshot1'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.snapshot_test.format(self.args['repository'], snap_name, 1, 30),
        )
        self.invoke_runner()
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        assert 1 == len(snapshot['snapshots'])
        assert snap_name == snapshot['snapshots'][0]['snapshot']

    def test_snapshot_datemath(self):
        self.create_indices(5)
        self.create_repository()
        snap_name = '<snapshot-{now-1d/d}>'
        _ = (datetime.now(timezone.utc) - timedelta(days=1)).strftime('%Y.%m.%d')
        snap_name_parsed = f"snapshot-{_}"
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.snapshot_test.format(self.args['repository'], snap_name, 1, 30),
        )
        self.invoke_runner()
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        assert 1 == len(snapshot['snapshots'])
        assert snap_name_parsed == snapshot['snapshots'][0]['snapshot']

    def test_snapshot_ignore_empty_list(self):
        self.create_indices(5)
        self.create_repository()
        snap_name = 'snapshot1'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.test_682.format(self.args['repository'], snap_name, True, 1, 30),
        )
        self.invoke_runner()
        snapshot = {'snapshots': []}
        try:
            snapshot = get_snapshot(self.client, self.args['repository'], '*')
        except FailedExecution:
            pass
        assert 0 == len(snapshot['snapshots'])
        assert 0 == len(get_indices(self.client))

    def test_snapshot_do_not_ignore_empty_list(self):
        self.create_indices(5)
        self.create_repository()
        snap_name = 'snapshot1'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.test_682.format(self.args['repository'], snap_name, False, 1, 30),
        )
        self.invoke_runner()
        snapshot = {'snapshots': []}
        try:
            snapshot = get_snapshot(self.client, self.args['repository'], '*')
        except FailedExecution:
            pass
        assert 0 == len(snapshot['snapshots'])
        assert 5 == len(get_indices(self.client))

    def test_no_repository(self):
        self.create_indices(5)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.snapshot_test.format(' ', 'snap_name', 1, 30),
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_extra_option(self):
        self.create_indices(5)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.bad_option_proto_test.format('snapshot')
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code


class TestCLISnapshot(CuratorTestCase):
    def test_snapshot(self):
        self.create_indices(5)
        self.create_repository()
        snap_name = 'snapshot1'
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'snapshot',
            '--repository',
            self.args['repository'],
            '--name',
            snap_name,
            '--wait_interval',
            '1',
            '--max_wait',
            '30',
            '--filter_list',
            '{"filtertype":"none"}',
        ]
        assert 0 == self.run_subprocess(args, logname='TestCLISnapshot.test_snapshot')
        snapshot = get_snapshot(self.client, self.args['repository'], '*')
        assert 1 == len(snapshot['snapshots'])
        assert snap_name == snapshot['snapshots'][0]['snapshot']



# -*- coding: utf-8 -*-
#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

# pylint: disable=redefined-builtin, invalid-name
import sys
import os
import re
from datetime import datetime

VERSIONFILE = '../curator/_version.py'
COPYRIGHT_YEARS = f'2011-{datetime.now().year}'

verstrline = open(VERSIONFILE, "rt", encoding='utf-8').read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError(f"Unable to find version string in {VERSIONFILE}.")

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('../'))

extensions = ['sphinx.ext.autodoc', 'sphinx.ext.doctest', 'sphinx.ext.intersphinx']

autoclass_content = "both"

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Elasticsearch Curator'
copyright = f'{COPYRIGHT_YEARS}, Elasticsearch'

release = verstr
version = '.'.join(release.split('.')[:2])

exclude_patterns = ['_build']

pygments_style = "sphinx"

on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3.13', None),
    'es_client': ('https://es-client.readthedocs.io/en/v9.0.0', None),
    'elasticsearch9': ('https://elasticsearch-py.readthedocs.io/en/v9.1.0', None),
    'voluptuous': ('http://alecthomas.github.io/voluptuous/docs/_build/html', None),
    'click': ('https://click.palletsprojects.com/en/stable', None),
}



"""Test date math with indices"""

# pylint: disable=C0115, C0116, invalid-name
from datetime import datetime, timedelta, timezone
from unittest import SkipTest
from curator.helpers.date_ops import parse_datemath

from . import CuratorTestCase


class TestParseDateMath(CuratorTestCase):
    colonblow = (
        'Date math expressions with colons are not supported in Elasticsearch '
        'versions 8.0 - 8.6'
    )

    def test_assorted_datemaths(self):
        now = datetime.now(timezone.utc)
        oneday = timedelta(days=1)
        ymd = '%Y.%m.%d'
        for test_string, expected in [
            (
                '<prefix-{now}-suffix>',
                f"prefix-{now.strftime(ymd)}-suffix",
            ),
            (
                '<prefix-{now-1d/d}>',
                f"prefix-{(now-oneday).strftime(ymd)}",
            ),
            (
                '<{now+1d/d}>',
                f"{(now+oneday).strftime(ymd)}",
            ),
            (
                '<{now+1d/d}>',
                f"{(now+oneday).strftime(ymd)}",
            ),
            (
                '<{now+10d/d{yyyy-MM}}>',
                f"{(now+timedelta(days=10)).strftime('%Y-%m')}",
            ),
        ]:
            assert expected == parse_datemath(self.client, test_string)

    def test_complex_datemath(self):
        v = self.get_version()
        if (8, 0, 0) < v < (8, 7, 0):
            raise SkipTest(f'{self.colonblow} (current version: {v})')
        now = datetime.now(timezone.utc)
        _ = now + timedelta(days=10) - timedelta(hours=7)
        test_string = '<{now+10d/h{yyyy-MM-dd-HH|-07:00}}>'
        expected = f"{_.strftime('%Y-%m-%d-%H')}"
        assert expected == parse_datemath(self.client, test_string)

    def test_very_complex_datemaths(self):
        v = self.get_version()
        if (8, 0, 0) < v < (8, 7, 0):
            raise SkipTest(f'{self.colonblow} (current version: {v})')
        test_string = '<.prefix-{2001-01-01-13||+1h/h{yyyy-MM-dd-HH|-07:00}}-suffix>'
        expected = '.prefix-2001-01-01-14-suffix'
        assert expected == parse_datemath(self.client, test_string)



"""Test shrink action"""

# pylint: disable=C0115, C0116, invalid-name, attribute-defined-outside-init
import os
import logging
from curator.helpers.getters import get_indices
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')

SHRINK = (
    '---\n'
    'actions:\n'
    '  1:\n'
    '    description: "Act on indices as filtered"\n'
    '    action: shrink\n'
    '    options:\n'
    '      shrink_node: {0}\n'
    '      node_filters:\n'
    '          {1}: {2}\n'
    '      number_of_shards: {3}\n'
    '      number_of_replicas: {4}\n'
    '      shrink_prefix: {5}\n'
    '      shrink_suffix: {6}\n'
    '      delete_after: {7}\n'
    '      wait_for_rebalance: {8}\n'
    '    filters:\n'
    '      - filtertype: pattern\n'
    '        kind: prefix\n'
    '        value: my\n'
)

NO_PERMIT_MASTERS = (
    '---\n'
    'actions:\n'
    '  1:\n'
    '    description: "Act on indices as filtered"\n'
    '    action: shrink\n'
    '    options:\n'
    '      shrink_node: {0}\n'
    '      number_of_shards: {1}\n'
    '      number_of_replicas: {2}\n'
    '      shrink_prefix: {3}\n'
    '      shrink_suffix: {4}\n'
    '      delete_after: {5}\n'
    '    filters:\n'
    '      - filtertype: pattern\n'
    '        kind: prefix\n'
    '        value: my\n'
)

WITH_EXTRA_SETTINGS = (
    '---\n'
    'actions:\n'
    '  1:\n'
    '    description: "Act on indices as filtered"\n'
    '    action: shrink\n'
    '    options:\n'
    '      shrink_node: {0}\n'
    '      node_filters:\n'
    '          {1}: {2}\n'
    '      number_of_shards: {3}\n'
    '      number_of_replicas: {4}\n'
    '      shrink_prefix: {5}\n'
    '      shrink_suffix: {6}\n'
    '      delete_after: {7}\n'
    '      extra_settings:\n'
    '        settings:\n'
    '          {8}: {9}\n'
    '        aliases:\n'
    '          my_alias: {10}\n'
    '      post_allocation:\n'
    '          allocation_type: {11}\n'
    '          key: {12}\n'
    '          value: {13}\n'
    '    filters:\n'
    '      - filtertype: pattern\n'
    '        kind: prefix\n'
    '        value: my\n'
)

COPY_ALIASES = (
    '---\n'
    'actions:\n'
    '  1:\n'
    '    description: "Act on indices as filtered"\n'
    '    action: shrink\n'
    '    options:\n'
    '      shrink_node: {0}\n'
    '      node_filters:\n'
    '          {1}: {2}\n'
    '      number_of_shards: {3}\n'
    '      number_of_replicas: {4}\n'
    '      shrink_prefix: {5}\n'
    '      shrink_suffix: {6}\n'
    '      copy_aliases: {7}\n'
    '      delete_after: {8}\n'
    '    filters:\n'
    '      - filtertype: pattern\n'
    '        kind: prefix\n'
    '        value: my\n'
)

SHRINK_FILTER_BY_SHARDS = (
    '---\n'
    'actions:\n'
    '  1:\n'
    '    description: "Act on indices as filtered"\n'
    '    action: shrink\n'
    '    options:\n'
    '      shrink_node: {0}\n'
    '      node_filters:\n'
    '          {1}: {2}\n'
    '      number_of_shards: {3}\n'
    '      number_of_replicas: {4}\n'
    '      shrink_prefix: {5}\n'
    '      shrink_suffix: {6}\n'
    '      delete_after: {7}\n'
    '      wait_for_rebalance: {8}\n'
    '    filters:\n'
    '      - filtertype: shards\n'
    '        number_of_shards: {9}\n'
    '        shard_filter_behavior: {10}\n'
)


class TestActionFileShrink(CuratorTestCase):
    def builder(self, action_args):
        self.idx = 'my_index'
        suffix = '-shrunken'
        self.target = f'{self.idx}{suffix}'
        self.create_index(self.idx, shards=2)
        self.add_docs(self.idx)
        # add alias in the source index
        self.alias = 'my_alias'
        alias_actions = []
        alias_actions.append({'add': {'index': self.idx, 'alias': self.alias}})
        self.client.indices.update_aliases(actions=alias_actions)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], action_args)
        self.invoke_runner()

    def test_shrink(self):
        suffix = '-shrunken'
        self.builder(
            SHRINK.format(
                'DETERMINISTIC',
                'permit_masters',
                'True',
                1,
                0,
                '',
                suffix,
                'True',
                'True',
            )
        )
        indices = get_indices(self.client)
        assert 1 == len(indices)
        assert indices[0] == self.target

    def test_permit_masters_false(self):
        suffix = '-shrunken'
        self.builder(
            NO_PERMIT_MASTERS.format('DETERMINISTIC', 1, 0, '', suffix, 'True')
        )
        indices = get_indices(self.client)
        assert 1 == len(indices)
        assert indices[0] == self.idx
        assert 1 == self.result.exit_code

    def test_shrink_non_green(self):
        suffix = '-shrunken'
        self.client.indices.create(
            index='just_another_index',
            settings={'number_of_shards': 1, 'number_of_replicas': 1},
        )
        self.builder(
            SHRINK.format(
                'DETERMINISTIC',
                'permit_masters',
                'True',
                1,
                0,
                '',
                suffix,
                'True',
                'True',
            )
        )
        indices = get_indices(self.client)
        assert 2 == len(indices)
        assert 1 == self.result.exit_code

    def test_shrink_with_extras(self):
        suffix = '-shrunken'
        allocation_type = 'exclude'
        key = '_name'
        value = 'not_this_node'
        self.builder(
            WITH_EXTRA_SETTINGS.format(
                'DETERMINISTIC',
                'permit_masters',
                'True',
                1,
                0,
                '',
                suffix,
                'False',
                'index.codec',
                'best_compression',
                dict(),
                allocation_type,
                key,
                value,
            )
        )
        indices = get_indices(self.client)
        assert 2 == len(indices)
        settings = self.client.indices.get_settings()
        assert (
            value
            == settings[self.target]['settings']['index']['routing']['allocation'][
                allocation_type
            ][key]
        )
        assert (
            ''
            == settings[self.idx]['settings']['index']['routing']['allocation'][
                'require'
            ]['_name']
        )
        assert 'best_compression' == settings[self.target]['settings']['index']['codec']
        # This was erroneously testing for True in previous releases. But the target
        # index should never have had the alias as `copy_aliases` was never set
        # to True for this test.

    def test_shrink_with_copy_alias(self):
        suffix = '-shrunken'
        self.builder(
            COPY_ALIASES.format(
                'DETERMINISTIC',
                'permit_masters',
                'True',
                1,
                0,
                '',
                suffix,
                'True',
                'True',
            )
        )
        indices = get_indices(self.client)
        assert 1 == len(indices)
        assert indices[0] == self.target
        assert self.client.indices.exists_alias(index=self.target, name=self.alias)

    def test_shrink_without_rebalance(self):
        suffix = '-shrunken'
        self.builder(
            SHRINK.format(
                'DETERMINISTIC',
                'permit_masters',
                'True',
                1,
                0,
                '',
                suffix,
                'True',
                'False',
            )
        )
        indices = get_indices(self.client)
        assert 1 == len(indices)
        assert indices[0] == self.target

    def test_shrink_implicit_shard_filter(self):
        self.create_index('my_invalid_shrink_index', shards=1)
        self.create_index('my_valid_shrink_index', shards=5)
        suffix = '-shrunken'
        self.builder(
            SHRINK.format(
                'DETERMINISTIC',
                'permit_masters',
                'True',
                1,
                0,
                '',
                suffix,
                'True',
                'False',
            )
        )

        indices = get_indices(self.client)
        assert 3 == len(indices)
        assert 'my_invalid_shrink_index-shrunken' not in indices
        assert 'my_valid_shrink_index-shrunken' in indices

    def test_shrink_explicit_shard_filter(self):
        self.create_index('my_invalid_shrink_index', shards=3)
        self.create_index('my_valid_shrink_index', shards=5)
        suffix = '-shrunken'
        self.builder(
            SHRINK_FILTER_BY_SHARDS.format(
                'DETERMINISTIC',
                'permit_masters',
                'True',
                1,
                0,
                '',
                suffix,
                'True',
                'False',
                5,
                'greater_than_or_equal',
            )
        )
        indices = get_indices(self.client)
        assert 3 == len(indices)
        assert 'my_invalid_shrink_index-shrunken' not in indices
        assert 'my_valid_shrink_index-shrunken' in indices
        assert 'my_index-shrunken' not in indices


class TestCLIShrink(CuratorTestCase):
    def builder(self):
        self.loogger = logging.getLogger('TestCLIShrink.builder')
        self.idx = 'my_index'
        self.suffix = '-shrunken'
        self.target = f'{self.idx}{self.suffix}'
        self.create_index(self.idx, shards=2)
        self.add_docs(self.idx)
        # add alias in the source index
        self.alias = 'my_alias'
        alias_actions = []
        alias_actions.append({'add': {'index': self.idx, 'alias': self.alias}})
        self.client.indices.update_aliases(actions=alias_actions)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.loogger.debug('Test pre-execution build phase complete.')

    def test_shrink(self):
        self.builder()
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'shrink',
            '--shrink_node',
            'DETERMINISTIC',
            '--node_filters',
            '{"permit_masters":"true"}',
            '--number_of_shards',
            "1",
            '--number_of_replicas',
            "0",
            '--shrink_suffix',
            self.suffix,
            '--delete_after',
            '--wait_for_rebalance',
            '--filter_list',
            '{"filtertype":"none"}',
        ]
        assert 0 == self.run_subprocess(args, logname='TestCLIShrink.test_shrink')
        indices = get_indices(self.client)
        assert 1 == len(indices)
        assert indices[0] == self.target



"""Test es_repo_mgr script and functions"""

# pylint: disable=C0115, C0116, invalid-name
import logging
import os
from click import testing as clicktest
from curator import repo_mgr_cli
from curator.helpers.testers import repository_exists
from . import CuratorTestCase
from . import testvars

LOGGER = logging.getLogger(__name__)
REPO_PATH = '/media'

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')

# class TestLoggingModules(CuratorTestCase):
#     def test_logger_without_null_handler(self):
#         from unittest.mock import patch, Mock
#         mock = Mock()
#         modules = {'logger': mock, 'logger.NullHandler': mock.module}
#         self.write_config(
#             self.args['configfile'],
#             testvars.client_conf_logfile.format(HOST, os.devnull)
#         )
#         with patch.dict('sys.modules', modules):
#             self.create_repository()
#             test = clicktest.CliRunner()
#             result = test.invoke(
#                 repo_mgr_cli,
#                 ['--config', self.args['configfile'], 'show']
#             )
#         self.assertEqual(self.args['repository'], result.output.rstrip())


class TestCLIRepositoryCreate(CuratorTestCase):
    def test_create_fs_repository_success(self):
        self.write_config(
            self.args['configfile'],
            testvars.client_conf_logfile.format(HOST, os.devnull),
        )
        test = clicktest.CliRunner()
        result = test.invoke(
            repo_mgr_cli,
            [
                '--config',
                self.args['configfile'],
                'create',
                'fs',
                '--name',
                self.args['repository'],
                '--location',
                REPO_PATH,
                '--verify',
            ],
        )
        assert 1 == len(
            self.client.snapshot.get_repository(name=self.args['repository'])
        )
        assert 0 == result.exit_code

    def test_create_fs_repository_fail(self):
        self.write_config(
            self.args['configfile'],
            testvars.client_conf_logfile.format(HOST, os.devnull),
        )
        test = clicktest.CliRunner()
        result = test.invoke(
            repo_mgr_cli,
            [
                '--config',
                self.args['configfile'],
                'create',
                'fs',
                '--name',
                self.args['repository'],
                '--location',
                os.devnull,
                '--verify',
            ],
        )
        assert 1 == result.exit_code

    def test_create_s3_repository_fail(self):
        self.write_config(
            self.args['configfile'],
            testvars.client_conf_logfile.format(HOST, os.devnull),
        )
        test = clicktest.CliRunner()
        result = test.invoke(
            repo_mgr_cli,
            [
                '--config',
                self.args['configfile'],
                'create',
                's3',
                '--bucket',
                'mybucket',
                '--name',
                self.args['repository'],
                '--verify',
            ],
        )
        assert 1 == result.exit_code

    def test_create_azure_repository_fail(self):
        self.write_config(
            self.args['configfile'],
            testvars.client_conf_logfile.format(HOST, os.devnull),
        )
        test = clicktest.CliRunner()
        result = test.invoke(
            repo_mgr_cli,
            [
                '--config',
                self.args['configfile'],
                'create',
                'azure',
                '--container',
                'mybucket',
                '--name',
                self.args['repository'],
                '--verify',
            ],
        )
        assert 1 == result.exit_code

    def test_create_gcs_repository_fail(self):
        self.write_config(
            self.args['configfile'],
            testvars.client_conf_logfile.format(HOST, os.devnull),
        )
        test = clicktest.CliRunner()
        result = test.invoke(
            repo_mgr_cli,
            [
                '--config',
                self.args['configfile'],
                'create',
                'gcs',
                '--bucket',
                'mybucket',
                '--name',
                self.args['repository'],
                '--verify',
            ],
        )
        assert 1 == result.exit_code


class TestCLIDeleteRepository(CuratorTestCase):
    def test_delete_repository_success(self):
        self.create_repository()
        self.write_config(
            self.args['configfile'],
            testvars.client_conf_logfile.format(HOST, os.devnull),
        )
        test = clicktest.CliRunner()
        _ = test.invoke(
            repo_mgr_cli,
            [
                '--config',
                self.args['configfile'],
                'delete',
                '--yes',  # This ensures no prompting will happen
                '--name',
                self.args['repository'],
            ],
        )
        assert not repository_exists(self.client, self.args['repository'])

    def test_delete_repository_notfound(self):
        self.write_config(
            self.args['configfile'],
            testvars.client_conf_logfile.format(HOST, os.devnull),
        )
        test = clicktest.CliRunner()
        result = test.invoke(
            repo_mgr_cli,
            [
                '--config',
                self.args['configfile'],
                'delete',
                '--yes',  # This ensures no prompting will happen
                '--name',
                self.args['repository'],
            ],
        )
        assert 1 == result.exit_code


class TestCLIShowRepositories(CuratorTestCase):
    def test_show_repository(self):
        self.create_repository()
        self.write_config(
            self.args['configfile'],
            testvars.client_conf_logfile.format(HOST, os.devnull),
        )
        test = clicktest.CliRunner()
        result = test.invoke(
            repo_mgr_cli, ['--config', self.args['configfile'], 'show']
        )
        # The splitlines()[-1] allows me to only capture the last line and ignore
        # other output
        assert (
            self.args['repository']
            == result.stdout_bytes.decode('utf-8').splitlines()[-1]
        )



"""Test integrations"""

# pylint: disable=C0115, C0116, invalid-name
import os
import warnings
import pytest
from elasticsearch9 import Elasticsearch
from elasticsearch9.exceptions import ElasticsearchWarning, NotFoundError
from curator.exceptions import ConfigurationError
from curator.helpers.getters import get_indices
from curator import IndexList
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')


class TestFilters(CuratorTestCase):
    def test_filter_by_alias(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.filter_by_alias.format('testalias', False)
        )
        self.create_index('my_index')
        self.create_index('dummy')
        self.client.indices.put_alias(index='dummy', name=alias)
        self.invoke_runner()
        assert 1 == len(get_indices(self.client))

    def test_filter_by_array_of_aliases(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.filter_by_alias.format(' [ testalias, foo ]', False),
        )
        self.create_index('my_index')
        self.create_index('dummy')
        self.client.indices.put_alias(index='dummy', name=alias)
        self.invoke_runner()
        assert 2 == len(get_indices(self.client))

    def test_filter_by_alias_bad_aliases(self):
        alias = 'testalias'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.filter_by_alias.format('{"this":"isadict"}', False),
        )
        self.create_index('my_index')
        self.create_index('dummy')
        self.client.indices.put_alias(index='dummy', name=alias)
        self.invoke_runner()
        assert isinstance(self.result.exception, ConfigurationError)
        assert 2 == len(get_indices(self.client))

    def test_filter_closed(self):
        idx1 = 'dummy'
        idx2 = 'my_index'
        ptrn = f'{idx1}*,{idx2}*'
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.filter_closed.format("True")
        )
        self.create_index(idx1)
        self.create_index(idx2)  # This one will be closed, ergo not deleted
        self.client.indices.close(index=idx2)
        self.invoke_runner()
        assert 1 == len(get_indices(self.client))
        result = self.client.indices.get(index=ptrn, expand_wildcards='open,closed')
        assert idx2 == list(dict(result).keys())[0]

    def test_field_stats_skips_empty_index(self):
        delete_field_stats = (
            '---\n'
            'actions:\n'
            '  1:\n'
            '    action: delete_indices\n'
            '    filters:\n'
            '      - filtertype: age\n'
            '        source: field_stats\n'
            '        direction: older\n'
            '        field: "{0}"\n'
            '        unit: days\n'
            '        unit_count: 1\n'
            '        stats_result: min_value\n'
        )
        idx = 'my_index'
        zero = 'zero'
        field = '@timestamp'
        time = '2017-12-31T23:59:59.999Z'
        # Create idx with a single, @timestamped doc
        self.client.create(index=idx, id=1, document={field: time})
        # Flush to ensure it's written
        # Decorators make this pylint exception necessary
        # pylint: disable=E1123
        self.client.indices.flush(index=idx, force=True)
        self.client.indices.refresh(index=idx)
        # Create zero with no docs
        self.create_index(zero)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(self.args['actionfile'], delete_field_stats.format(field))
        self.invoke_runner()
        # It should skip deleting 'zero', as it has 0 docs
        assert [zero] == get_indices(self.client)


class TestIndexList(CuratorTestCase):
    """Test some of the IndexList particulars using a live ES instance/cluster"""

    IDX1 = 'dummy1'
    IDX2 = 'dummy2'
    IDX3 = 'my_index'

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog):
        # pylint: disable=attribute-defined-outside-init
        self._caplog = caplog

    def test_get_index_stats_with_404(self):
        """
        Check to ensure that index_stats are being collected if one index is missing
        """
        # expected = f'Index was initiallly present, but now is not: {self.IDX2}'
        self.create_index(self.IDX1)
        self.create_index(self.IDX2)
        self.create_index(self.IDX3)
        ilo = IndexList(self.client)
        assert ilo.indices == [self.IDX1, self.IDX2, self.IDX3]
        self.client.indices.delete(index=f'{self.IDX1},{self.IDX2}')
        ilo.get_index_stats()
        # with self._caplog.at_level(logging.WARNING):
        #     ilo.get_index_stats()
        #     # Guarantee we're getting the expected WARNING level message
        #     assert self._caplog.records[-1].message == expected
        assert ilo.indices == [self.IDX3]

    def test_get_index_state(self):
        """Check to ensure that open/close status is properly being recognized"""
        self.create_index(self.IDX1)
        self.create_index(self.IDX2)
        ilo = IndexList(self.client)
        ilo.get_index_state()
        assert ilo.indices == [self.IDX1, self.IDX2]
        assert ilo.index_info[self.IDX1]['state'] == 'open'
        assert ilo.index_info[self.IDX2]['state'] == 'open'
        # ElasticsearchWarning: the default value for the wait_for_active_shards
        # parameter will change from '0' to 'index-setting' in version 8;
        # specify 'wait_for_active_shards=index-setting' to adopt the future default
        # behaviour, or 'wait_for_active_shards=0' to preserve today's behaviour
        warnings.filterwarnings("ignore", category=ElasticsearchWarning)
        self.client.indices.close(index=self.IDX2)
        ilo.get_index_state()
        assert ilo.index_info[self.IDX2]['state'] == 'close'

    def test_get_index_state_alias(self):
        """Check to ensure that open/close status catches an alias"""
        alias = {self.IDX2: {}}
        self.create_index(self.IDX1)
        self.create_index(self.IDX2)
        ilo = IndexList(self.client)
        assert ilo.indices == [self.IDX1, self.IDX2]
        self.client.indices.delete(index=self.IDX2)
        self.client.indices.create(index=self.IDX3, aliases=alias)
        ilo.get_index_state()
        assert ilo.indices == [self.IDX1, self.IDX3]

    def test_population_check_missing_index(self):
        """
        If index_info is missing an index, test to ensure it is populated with
        the zero value
        """
        key = 'docs'
        self.create_index(self.IDX1)
        ilo = IndexList(self.client)
        ilo.population_check(self.IDX1, key)
        assert ilo.index_info[self.IDX1][key] == 0

    def test_population_check_missing_key(self):
        """
        If index_info is missing a key, test to ensure it is populated with a
        zero value
        """
        key = 'docs'
        self.create_index(self.IDX1)
        ilo = IndexList(self.client)
        ilo.get_index_stats()
        assert ilo.indices == [self.IDX1]
        assert ilo.index_info[self.IDX1][key] == 0
        del ilo.index_info[self.IDX1][key]
        ilo.population_check(self.IDX1, key)
        assert ilo.index_info[self.IDX1][key] == 0

    def test_not_needful(self):
        """
        Check if get_index_stats can be skipped if already populated

        THIS IS LITERALLY FOR CODE COVERAGE, so a ``continue`` line in the
        function is tested.
        """
        key = 'docs'
        self.create_index(self.IDX1)
        ilo = IndexList(self.client)
        ilo.get_index_stats()
        ilo.get_index_stats()  # index_info is already populated and it will skip
        assert ilo.index_info[self.IDX1][key] == 0

    def test_search_pattern_1(self):
        """Check to see if providing a search_pattern limits the index list"""
        pattern = 'd*'
        self.create_index(self.IDX1)
        self.create_index(self.IDX2)
        self.create_index(self.IDX3)
        ilo1 = IndexList(self.client)
        ilo2 = IndexList(self.client, search_pattern=pattern)
        assert ilo1.indices == [self.IDX1, self.IDX2, self.IDX3]
        assert ilo2.indices == [self.IDX1, self.IDX2]


def cleanup(client, idx):
    """Cleanup indices"""
    try:
        client.indices.delete(index=idx)
    except NotFoundError:
        pass


@pytest.mark.parametrize(
    'idx, pattern, hidden, response',
    [
        ('my_index', 'my_*', True, ['my_index']),
        ('my_index', 'my_*', False, []),
        ('my_index', '*_index', True, ['my_index']),
        ('my_index', '*_index', False, []),
        ('.my_index', '.my_*', True, ['.my_index']),
        # ('.my_index', '.my_*', False, []),  # This is a weird behavior or a bug.
        # Check in test__bug__include_hidden_dot_prefix_index
        ('.my_index', '*_index', True, ['.my_index']),
        ('.my_index', '*_index', False, []),
    ],
)
def test_include_hidden_index(idx, pattern, hidden, response):
    """Test that a hidden index is included when include_hidden is True"""
    host = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')
    client = Elasticsearch(hosts=host, request_timeout=300)
    warnings.filterwarnings("ignore", category=ElasticsearchWarning)
    cleanup(client, idx)
    client.indices.create(index=idx)
    client.indices.put_settings(index=idx, body={'index': {'hidden': True}})
    ilo = IndexList(client, search_pattern=pattern, include_hidden=hidden)
    assert ilo.indices == response
    # Manual teardown because hidden indices are not returned by get_indices
    del ilo
    cleanup(client, idx)


@pytest.mark.parametrize(
    'idx, pattern, hidden, response',
    [
        ('.my_index', '.my_*', False, ['.my_index']),
        ('.your_index', '.your_*', False, ['.your_index']),
    ],
)
def test__bug__include_hidden_dot_prefix_index(idx, pattern, hidden, response):
    """
    Test that a hidden index is included when include_hidden is True when the
    multi-target search pattern starts with a leading dot and includes a wildcard,
    e.g. '.my_*' or '.your_*'

    This is a bug, and the test is to confirm that behavior until it's fixed.

    https://github.com/elastic/elasticsearch/issues/124167
    """
    host = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')
    client = Elasticsearch(hosts=host, request_timeout=300)
    warnings.filterwarnings("ignore", category=ElasticsearchWarning)
    cleanup(client, idx)
    client.indices.create(index=idx)
    client.indices.put_settings(index=idx, body={'index': {'hidden': True}})
    hidden_test = client.indices.get_settings(
        index=idx, filter_path=f'\\{idx}.settings.index.hidden', expand_wildcards='all'
    )
    # We're confirming that our index that starts with a leading dot is hidden
    assert hidden_test[idx]['settings']['index']['hidden'] == 'true'
    ilo = IndexList(client, search_pattern=pattern, include_hidden=hidden)
    # And now we assert that the index is still included in the list
    # WHICH SHOULD NOT HAPPEN -- THIS IS A BUG
    assert ilo.indices == response
    # Manual teardown because hidden indices are not returned by get_indices
    del ilo
    cleanup(client, idx)



"""test_action_forcemerge"""
# pylint: disable=missing-function-docstring, missing-class-docstring, protected-access, attribute-defined-outside-init
from unittest import TestCase
from unittest.mock import Mock
from curator.actions import ForceMerge
from curator.exceptions import FailedExecution, MissingArgument
from curator import IndexList
# Get test variables and constants from a single source
from . import testvars

class TestActionForceMerge(TestCase):
    VERSION = {'version': {'number': '8.0.0'} }
    def builder(self):
        self.client = Mock()
        self.client.info.return_value = self.VERSION
        self.client.cat.indices.return_value = testvars.state_one
        self.client.indices.get_settings.return_value = testvars.settings_one
        self.client.indices.stats.return_value = testvars.stats_one
        self.client.indices.exists_alias.return_value = False
        self.client.indices.segments.return_value = testvars.shards
        self.ilo = IndexList(self.client)
    def test_init_raise_bad_client(self):
        self.assertRaises(
            TypeError, ForceMerge, 'invalid', max_num_segments=2)
    def test_init_raise_no_segment_count(self):
        self.builder()
        self.assertRaises(MissingArgument, ForceMerge, self.ilo)
    def test_init(self):
        self.builder()
        fmo = ForceMerge(self.ilo, max_num_segments=2)
        self.assertEqual(self.ilo, fmo.index_list)
        self.assertEqual(self.client, fmo.client)
    def test_do_dry_run(self):
        self.builder()
        self.client.indices.forcemerge.return_value = None
        self.client.indices.optimize.return_value = None
        fmo = ForceMerge(self.ilo, max_num_segments=2)
        self.assertIsNone(fmo.do_dry_run())
    def test_do_action(self):
        self.builder()
        self.client.indices.forcemerge.return_value = None
        fmo = ForceMerge(self.ilo, max_num_segments=2)
        self.assertIsNone(fmo.do_action())
    def test_do_action_with_delay(self):
        self.builder()
        self.client.indices.forcemerge.return_value = None
        fmo = ForceMerge(self.ilo, max_num_segments=2, delay=0.050)
        self.assertIsNone(fmo.do_action())
    def test_do_action_raises_exception(self):
        self.builder()
        self.client.indices.forcemerge.return_value = None
        self.client.indices.optimize.return_value = None
        self.client.indices.forcemerge.side_effect = testvars.fake_fail
        self.client.indices.optimize.side_effect = testvars.fake_fail
        fmo = ForceMerge(self.ilo, max_num_segments=2)
        self.assertRaises(FailedExecution, fmo.do_action)



"""Test filter and schema validators"""

from unittest import TestCase
from voluptuous import Schema
from es_client.exceptions import FailedValidation
from es_client.schemacheck import SchemaCheck
from curator.exceptions import ConfigurationError
from curator.validators.filter_functions import validfilters, singlefilter


def shared_result(config, action):
    """Test functions should be able to reuse this"""
    return SchemaCheck(
        config, Schema(validfilters(action)), 'filters', 'testing'
    ).result()


class TestFilters(TestCase):
    def test_single_raises_configuration_error(self):
        data = {'max_num_segments': 1, 'exclude': True}
        self.assertRaises(ConfigurationError, singlefilter, 'forcemerge', data)


class TestFilterTypes(TestCase):
    def test_alias(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'alias',
                'aliases': ['alias1', 'alias2'],
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_age(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'age',
                'direction': 'older',
                'unit': 'days',
                'unit_count': 1,
                'source': 'field_stats',
                'field': '@timestamp',
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_age_with_string_unit_count(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'age',
                'direction': 'older',
                'unit': 'days',
                'unit_count': "1",
                'source': 'field_stats',
                'field': '@timestamp',
            }
        ]
        result = shared_result(config, action)
        self.assertEqual(1, result[0]['unit_count'])

    def test_allocated(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'allocated',
                'key': 'foo',
                'value': 'bar',
                'allocation_type': 'require',
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_closed(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'closed',
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_count(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'count',
                'count': 1,
                'reverse': True,
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_forcemerged(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'forcemerged',
                'max_num_segments': 1,
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_kibana(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'kibana',
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_opened(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'opened',
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_space_name_age(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'space',
                'disk_space': 1,
                'use_age': True,
                'exclude': False,
                'source': 'name',
                'timestring': '%Y.%m.%d',
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_space_name_age_string_float(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'space',
                'disk_space': "1.0",
                'use_age': True,
                'exclude': False,
                'source': 'name',
                'timestring': '%Y.%m.%d',
            }
        ]
        result = shared_result(config, action)
        self.assertEqual(1.0, result[0]['disk_space'])

    def test_space_name_age_no_ts(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'space',
                'disk_space': 1,
                'use_age': True,
                'exclude': False,
                'source': 'name',
            }
        ]
        schema = SchemaCheck(config, Schema(validfilters(action)), 'filters', 'testing')
        self.assertRaises(FailedValidation, schema.result)

    def test_space_field_stats_age(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'space',
                'disk_space': 1,
                'use_age': True,
                'exclude': False,
                'source': 'field_stats',
                'field': '@timestamp',
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_space_field_stats_age_no_field(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'space',
                'disk_space': 1,
                'use_age': True,
                'exclude': False,
                'source': 'field_stats',
            }
        ]
        schema = SchemaCheck(config, Schema(validfilters(action)), 'filters', 'testing')
        self.assertRaises(FailedValidation, schema.result)

    def test_space_creation_date_age(self):
        action = 'delete_indices'
        config = [
            {
                'filtertype': 'space',
                'disk_space': 1,
                'use_age': True,
                'exclude': False,
                'source': 'creation_date',
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_state(self):
        action = 'delete_snapshots'
        config = [
            {
                'filtertype': 'state',
                'state': 'SUCCESS',
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))

    def test_shards(self):
        action = 'shrink'
        config = [
            {
                'filtertype': 'shards',
                'number_of_shards': 5,
                'shard_filter_behavior': 'greater_than',
                'exclude': False,
            }
        ]
        self.assertEqual(config, shared_result(config, action))



"""Unit testing for helpers.creators functions"""

from unittest import TestCase
from unittest.mock import Mock
import pytest
from elastic_transport import ApiResponseMeta
from elasticsearch9 import NotFoundError, TransportError
from curator.exceptions import CuratorException, FailedExecution, MissingArgument
from curator.helpers import getters

FAKE_FAIL = Exception('Simulated Failure')
NAMED_INDICES = ["index-2015.01.01", "index-2015.02.01"]
REPO_NAME = 'repo_name'
TEST_REPO = {REPO_NAME: {}}
SNAP_NAME = 'snap_name'
SINGLE = {'snapshot': SNAP_NAME, 'indices': NAMED_INDICES}
SNAPSHOT = {'snapshots': [SINGLE]}
SNAPSHOTS = {
    'snapshots': [SINGLE, {'snapshot': 'snapshot-2015.03.01', 'indices': NAMED_INDICES}]
}


class TestByteSize(TestCase):
    """TestByteSize

    Test helpers.getters.byte_size functionality.
    """

    def test_byte_size(self):
        """test_byte_size

        Output should match expected
        """
        size = 3 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024
        unit = ['Z', 'E', 'P', 'T', 'G', 'M', 'K', '']
        for i in range(0, 7):
            assert f'3.0{unit[i]}B' == getters.byte_size(size)
            size /= 1024

    def test_byte_size_yotta(self):
        """test_byte_size_yotta

        Output should match expected
        """
        size = 3 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024
        assert '3.0YB' == getters.byte_size(size)

    def test_raise_invalid(self):
        """test_raise_invalid

        Should raise a TypeError exception if an invalid value is passed
        """
        with pytest.raises(TypeError):
            getters.byte_size('invalid')


class TestGetIndices(TestCase):
    """TestGetIndices

    Test helpers.getters.get_indices functionality.
    """

    IDX1 = 'index-2016.03.03'
    IDX2 = 'index-2016.03.04'
    RESPONSE = [{'index': IDX1, 'state': 'open'}, {'index': IDX2, 'state': 'open'}]

    def test_client_exception(self):
        """test_client_exception

        Should raise a FailedExecution exception when an upstream exception occurs
        """
        client = Mock()
        client.cat.indices.return_value = self.RESPONSE
        client.cat.indices.side_effect = FAKE_FAIL
        with pytest.raises(FailedExecution):
            getters.get_indices(client)

    def test_positive(self):
        """test_positive

        Output should match expected
        """
        client = Mock()
        client.cat.indices.return_value = self.RESPONSE
        self.assertEqual([self.IDX1, self.IDX2], sorted(getters.get_indices(client)))

    def test_empty(self):
        """test_empty

        Output should be an empty list
        """
        client = Mock()
        client.cat.indices.return_value = {}
        self.assertEqual([], getters.get_indices(client))


class TestGetRepository(TestCase):
    """TestGetRepository

    Test helpers.getters.get_repository functionality.
    """

    MULTI = {'other': {}, REPO_NAME: {}}

    def test_get_repository_missing_arg(self):
        """test_get_repository_missing_arg

        Should return an empty response if no repository name provided
        """
        client = Mock()
        client.snapshot.get_repository.return_value = {}
        assert not getters.get_repository(client)

    def test_get_repository_positive(self):
        """test_get_repository_positive

        Return value should match expected
        """
        client = Mock()
        client.snapshot.get_repository.return_value = TEST_REPO
        assert TEST_REPO == getters.get_repository(client, repository=REPO_NAME)

    def test_get_repository_transporterror_negative(self):
        """test_get_repository_transporterror_negative

        Should raise a CuratorException if a TransportError is raised first
        """
        client = Mock()
        client.snapshot.get_repository.side_effect = TransportError(
            503, ('exception', 'reason')
        )
        with pytest.raises(CuratorException, match=r'503 Check Elasticsearch logs'):
            getters.get_repository(client, repository=REPO_NAME)

    def test_get_repository_notfounderror_negative(self):
        """test_get_repository_notfounderror_negative

        Should raise a CuratorException if a NotFoundError is raised first
        """
        client = Mock()
        # 5 positional args for meta: status, http_version, headers, duration, node
        meta = ApiResponseMeta(404, '1.1', {}, 0.01, None)
        body = 'simulated error'
        msg = 'simulated error'
        # 3 positional args for NotFoundError: message, meta, body
        effect = NotFoundError(msg, meta, body)
        client.snapshot.get_repository.side_effect = effect
        with pytest.raises(CuratorException, match=r'Error: NotFoundError'):
            getters.get_repository(client, repository=REPO_NAME)

    def test_get_repository_all_positive(self):
        """test_get_repository_all_positive

        Return value should match expected with multiple repositories
        """
        client = Mock()
        client.snapshot.get_repository.return_value = self.MULTI
        assert self.MULTI == getters.get_repository(client)


class TestGetSnapshot(TestCase):
    """TestGetSnapshot

    Test helpers.getters.get_snapshot functionality.
    """

    def test_get_snapshot_missing_repository_arg(self):
        """test_get_snapshot_missing_repository_arg

        Should raise a MissingArgument exception when repository not passed
        """
        client = Mock()
        with pytest.raises(
            MissingArgument, match=r'No value for "repository" provided'
        ):
            getters.get_snapshot(client, snapshot=SNAP_NAME)

    def test_get_snapshot_positive(self):
        """test_get_snapshot_positive

        Output should match expected
        """
        client = Mock()
        client.snapshot.get.return_value = SNAPSHOT
        assert SNAPSHOT == getters.get_snapshot(
            client, repository=REPO_NAME, snapshot=SNAP_NAME
        )

    def test_get_snapshot_transporterror_negative(self):
        """test_get_snapshot_transporterror_negative

        Should raise a FailedExecution exception if a TransportError is raised first
        """
        client = Mock()
        client.snapshot.get_repository.return_value = TEST_REPO
        client.snapshot.get.side_effect = TransportError(401, "simulated error")
        with pytest.raises(FailedExecution, match=r'Error: 401'):
            getters.get_snapshot(client, repository=REPO_NAME, snapshot=SNAP_NAME)

    def test_get_snapshot_notfounderror_negative(self):
        """test_get_snapshot_notfounderror_negative

        Should raise a FailedExecution exception if a NotFoundError is raised first
        """
        client = Mock()
        client.snapshot.get_repository.return_value = TEST_REPO
        # 5 positional args for meta: status, http_version, headers, duration, node
        meta = ApiResponseMeta(404, '1.1', {}, 1.0, None)
        client.snapshot.get.side_effect = NotFoundError(
            'simulated error', meta, 'simulated error'
        )
        with pytest.raises(FailedExecution, match=r'Error: NotFoundError'):
            getters.get_snapshot(client, repository=REPO_NAME, snapshot=SNAP_NAME)


class TestGetSnapshotData(TestCase):
    """TestGetSnapshotData

    Test helpers.getters.get_snapshot_data functionality.
    """

    def test_missing_repo_arg(self):
        """test_missing_repo_arg

        Should raise a MissingArgument exception if the repository arg is missing
        """
        client = Mock()
        with pytest.raises(
            MissingArgument, match=r'No value for "repository" provided'
        ):
            getters.get_snapshot_data(client)

    def test_return_data(self):
        """test_return_data

        Output should match expected
        """
        client = Mock()
        client.snapshot.get.return_value = SNAPSHOTS
        client.snapshot.get_repository.return_value = TEST_REPO
        assert SNAPSHOTS['snapshots'] == getters.get_snapshot_data(
            client, repository=REPO_NAME
        )

    def test_raises_exception_onfail(self):
        """test_raises_exception_onfail

        Should raise a FailedExecution exception if a TransportError is raised upstream
        first
        """
        client = Mock()
        client.snapshot.get.return_value = SNAPSHOTS
        client.snapshot.get.side_effect = TransportError(401, "simulated error")
        client.snapshot.get_repository.return_value = TEST_REPO
        with pytest.raises(FailedExecution, match=r'Error: 401'):
            getters.get_snapshot_data(client, repository=REPO_NAME)


class TestNodeRoles(TestCase):
    """TestNodeRoles

    Test helpers.getters.node_roles functionality.
    """

    def test_node_roles(self):
        """test_node_roles

        Output should match expected
        """
        node_id = 'my_node'
        expected = ['data']
        client = Mock()
        client.nodes.info.return_value = {'nodes': {node_id: {'roles': expected}}}
        assert expected == getters.node_roles(client, node_id)


class TestSingleDataPath(TestCase):
    """TestSingleDataPath

    Test helpers.getters.single_data_path functionality.
    """

    def test_single_data_path(self):
        """test_single_data_path

        Return value should be True with only one data path
        """
        node_id = 'my_node'
        client = Mock()
        client.nodes.stats.return_value = {
            'nodes': {node_id: {'fs': {'data': ['one']}}}
        }
        assert getters.single_data_path(client, node_id)

    def test_two_data_paths(self):
        """test_two_data_paths

        Return value should be False with two data paths
        """
        node_id = 'my_node'
        client = Mock()
        client.nodes.stats.return_value = {
            'nodes': {node_id: {'fs': {'data': ['one', 'two']}}}
        }
        assert not getters.single_data_path(client, node_id)


class TestNameToNodeId(TestCase):
    """TestNameToNodeId

    Test helpers.getters.name_to_node_id functionality.
    """

    def test_positive(self):
        """test_positive

        Output should match expected
        """
        node_id = 'node_id'
        node_name = 'node_name'
        client = Mock()
        client.nodes.info.return_value = {'nodes': {node_id: {'name': node_name}}}
        assert node_id == getters.name_to_node_id(client, node_name)

    def test_negative(self):
        """test_negative

        Output should be None due to mismatch
        """
        node_id = 'node_id'
        node_name = 'node_name'
        client = Mock()
        client.nodes.info.return_value = {'nodes': {node_id: {'name': node_name}}}
        assert None is getters.name_to_node_id(client, 'wrong_name')


class TestNodeIdToName(TestCase):
    """TestNodeIdToName

    Test helpers.getters.node_id_to_name functionality.
    """

    def test_negative(self):
        """test_negative

        Output should be None due to mismatch
        """
        client = Mock()
        client.nodes.info.return_value = {
            'nodes': {'my_node_id': {'name': 'my_node_name'}}
        }
        assert None is getters.node_id_to_name(client, 'not_my_node_id')


class TestGetAliasActions(TestCase):
    """TestGetAliasActions

    Test helpers.getters.get_alias_actions functionality.
    """

    def test_get_alias_actions(self):
        """test_get_alias_actions"""
        name = 'alias1'
        aliases = {name: {}}
        oldidx = 'old'
        newidx = 'new'
        expected = [
            {'remove': {'index': oldidx, 'alias': name}},
            {'add': {'index': newidx, 'alias': name}},
        ]
        assert getters.get_alias_actions(oldidx, newidx, aliases) == expected


class TestGetTierPreference(TestCase):
    """TestGetTierPreference

    Test helpers.getters.get_tier_preference functionality.
    """

    def test_get_tier_preference1(self):
        """test_get_tier_preference1"""
        client = Mock()
        roles = ['data_cold', 'data_frozen', 'data_hot', 'data_warm']
        client.nodes.info.return_value = {'nodes': {'nodename': {'roles': roles}}}
        assert getters.get_tier_preference(client) == 'data_frozen'

    def test_get_tier_preference2(self):
        """test_get_tier_preference2"""
        client = Mock()
        roles = ['data_cold', 'data_hot', 'data_warm']
        client.nodes.info.return_value = {'nodes': {'nodename': {'roles': roles}}}
        assert getters.get_tier_preference(client) == 'data_cold,data_warm,data_hot'

    def test_get_tier_preference3(self):
        """test_get_tier_preference3"""
        client = Mock()
        roles = ['data_content']
        client.nodes.info.return_value = {'nodes': {'nodename': {'roles': roles}}}
        assert getters.get_tier_preference(client) == 'data_content'

    def test_get_tier_preference4(self):
        """test_get_tier_preference4"""
        client = Mock()
        roles = ['data_cold', 'data_frozen', 'data_hot', 'data_warm']
        client.nodes.info.return_value = {'nodes': {'nodename': {'roles': roles}}}
        assert (
            getters.get_tier_preference(client, target_tier='data_cold')
            == 'data_cold,data_warm,data_hot'
        )

    def test_get_tier_preference5(self):
        """test_get_tier_preference5"""
        client = Mock()
        roles = ['data_content']
        client.nodes.info.return_value = {'nodes': {'nodename': {'roles': roles}}}
        assert (
            getters.get_tier_preference(client, target_tier='data_hot')
            == 'data_content'
        )



"""Test index opening functionality"""

# pylint: disable=C0115, C0116, invalid-name
import os
import warnings
from elasticsearch9.exceptions import ElasticsearchWarning
from . import CuratorTestCase
from . import testvars

HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')
MET = 'metadata'


class TestActionFileOpenClosed(CuratorTestCase):
    def test_open_closed(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.optionless_proto.format('open')
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        # ElasticsearchWarning: the default value for the wait_for_active_shards
        # parameter will change from '0' to 'index-setting' in version 8;
        # specify 'wait_for_active_shards=index-setting' to adopt the future default
        # behaviour, or 'wait_for_active_shards=0' to preserve today's behaviour
        warnings.filterwarnings("ignore", category=ElasticsearchWarning)
        self.client.indices.close(index=idx2, ignore_unavailable=True)
        self.invoke_runner()
        check = {idx1: False, idx2: False}
        success = False
        csi = self.client.cluster.state(metric=MET)[MET]['indices']
        for idx in (idx1, idx2):
            if 'close' != csi[idx]['state']:
                check[idx] = True
        success = all(check.values())
        assert success

    def test_extra_option(self):
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'], testvars.bad_option_proto_test.format('open')
        )
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        # ElasticsearchWarning: the default value for the wait_for_active_shards
        # parameter will change from '0' to 'index-setting' in version 8;
        # specify 'wait_for_active_shards=index-setting' to adopt the future default
        # behaviour, or 'wait_for_active_shards=0' to preserve today's behaviour
        warnings.filterwarnings("ignore", category=ElasticsearchWarning)
        self.client.indices.close(index=idx2, ignore_unavailable=True)
        self.invoke_runner()
        csi = self.client.cluster.state(metric=MET)[MET]['indices']
        assert 'close' != csi[idx1]['state']
        assert 'close' == csi[idx2]['state']
        assert 1 == self.result.exit_code


class TestCLIOpenClosed(CuratorTestCase):
    def test_open_closed(self):
        idx1, idx2 = ('dummy', 'my_index')
        self.create_index(idx1)
        self.create_index(idx2)
        # ElasticsearchWarning: the default value for the wait_for_active_shards
        # parameter will change from '0' to 'index-setting' in version 8;
        # specify 'wait_for_active_shards=index-setting' to adopt the future default
        # behaviour, or 'wait_for_active_shards=0' to preserve today's behaviour
        warnings.filterwarnings("ignore", category=ElasticsearchWarning)
        self.client.indices.close(index=idx2, ignore_unavailable=True)
        args = self.get_runner_args()
        args += [
            '--config',
            self.args['configfile'],
            'open',
            '--filter_list',
            '{"filtertype":"pattern","kind":"prefix","value":"my"}',
        ]
        assert 0 == self.run_subprocess(
            args, logname='TestCLIOpenClosed.test_open_closed'
        )
        csi = self.client.cluster.state(metric=MET)[MET]['indices']
        for idx in (idx1, idx2):
            assert 'close' != csi[idx]['state']



"""Snapshot and Restore action classes"""

import logging
import re
from es_client.utils import ensure_list
from curator.helpers.date_ops import parse_datemath, parse_date_pattern
from curator.helpers.getters import get_indices
from curator.helpers.testers import (
    repository_exists,
    snapshot_running,
    verify_index_list,
    verify_repository,
    verify_snapshot_list,
)
from curator.helpers.utils import report_failure, to_csv, multitarget_match
from curator.helpers.waiters import wait_for_it

# pylint: disable=broad-except
from curator.exceptions import (
    ActionError,
    CuratorException,
    FailedRestore,
    FailedSnapshot,
    MissingArgument,
    SnapshotInProgress,
)


class Snapshot(object):
    """Snapshot Action Class

    Read more about identically named settings at:
    :py:meth:`elasticsearch.client.SnapshotClient.create`
    """

    def __init__(
        self,
        ilo,
        repository=None,
        name=None,
        ignore_unavailable=False,
        include_global_state=True,
        partial=False,
        wait_for_completion=True,
        wait_interval=9,
        max_wait=-1,
        skip_repo_fs_check=True,
    ):
        """
        :param ilo: An IndexList Object
        :param repository: Repository name.
        :param name: Snapshot name.
        :param ignore_unavailable: Ignore unavailable shards/indices.
        :param include_global_state: Store cluster global state with snapshot.
        :param partial: Do not fail if primary shard is unavailable.
        :param wait_for_completion: Wait for completion before returning.
        :param wait_interval: Seconds to wait between completion checks.
        :param max_wait: Maximum number of seconds to ``wait_for_completion``
        :param skip_repo_fs_check: Do not validate write access to repository on
            all cluster nodes before proceeding. Useful for shared filesystems
            where intermittent timeouts can affect validation, but won't likely
            affect snapshot success. (Default: ``True``)

        :type ilo: :py:class:`~.curator.indexlist.IndexList`
        :type repository: str
        :type name: str
        :type ignore_unavailable: bool
        :type include_global_state: bool
        :type partial: bool
        :type wait_for_completion: bool
        :type wait_interval: int
        :type max_wait: int
        :type skip_repo_fs_check: bool
        """
        verify_index_list(ilo)
        # Check here and don't bother with the rest of this if there are no
        # indices in the index list.
        ilo.empty_list_check()
        if not repository_exists(ilo.client, repository=repository):
            raise ActionError(
                f'Cannot snapshot indices to missing repository: {repository}'
            )
        if not name:
            raise MissingArgument('No value for "name" provided.')
        #: The :py:class:`~.curator.indexlist.IndexList` object passed from param
        #: ``ilo``
        self.index_list = ilo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`index_list`
        self.client = ilo.client
        #: The :py:func:`~.curator.helpers.date_ops.parse_date_pattern` rendered
        #: version of what was passed by param ``name``.
        self.name = parse_datemath(self.client, parse_date_pattern(name))
        #: Object attribute that gets the value of param ``repository``.
        self.repository = repository
        #: Object attribute that gets the value of param ``wait_for_completion``.
        self.wait_for_completion = wait_for_completion
        #: Object attribute that gets the value of param ``wait_interval``.
        self.wait_interval = wait_interval
        #: Object attribute that gets the value of param ``max_wait``.
        self.max_wait = max_wait
        #: Object attribute that gets the value of param ``skip_repo_fs_check``.
        self.skip_repo_fs_check = skip_repo_fs_check
        #: Object attribute that tracks the snapshot state.
        self.state = None
        #: Object attribute that contains the :py:func:`~.curator.helpers.utils.to_csv`
        #: output of the indices in :py:attr:`index_list`.
        self.indices = to_csv(ilo.indices)
        #: Object attribute that gets the value of param ``ignore_unavailable``.
        self.ignore_unavailable = ignore_unavailable
        #: Object attribute that gets the value of param ``include_global_state``.
        self.include_global_state = include_global_state
        #: Object attribute that gets the value of param ``partial``.
        self.partial = partial
        #: Object attribute dictionary compiled from :py:attr:`indices`,
        #: :py:attr:`ignore_unavailable`, :py:attr:`include_global_state`, and
        #: :py:attr:`partial`
        self.settings = {
            'indices': ilo.indices,
            'ignore_unavailable': self.ignore_unavailable,
            'include_global_state': self.include_global_state,
            'partial': self.partial,
        }

        self.loggit = logging.getLogger('curator.actions.snapshot')

    def get_state(self):
        """Get the state of the snapshot and set :py:attr:`state`"""
        try:
            self.state = self.client.snapshot.get(
                repository=self.repository, snapshot=self.name
            )['snapshots'][0]['state']
            return self.state
        except IndexError as exc:
            raise CuratorException(
                f'Snapshot "{self.name}" not found in repository "{self.repository}"'
            ) from exc

    def report_state(self):
        """
        Log the :py:attr:`state` of the snapshot and raise :py:exc:`FailedSnapshot` if
        :py:attr:`state` is not ``SUCCESS``
        """
        self.get_state()
        if self.state == 'SUCCESS':
            self.loggit.info('Snapshot %s successfully completed.', self.name)
        else:
            msg = f'Snapshot {self.name} completed with state: {self.state}'
            self.loggit.error(msg)
            raise FailedSnapshot(msg)

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        msg = (
            f'DRY-RUN: snapshot: {self.name} in repository {self.repository} '
            f'with arguments: {self.settings}'
        )
        self.loggit.info(msg)

    def do_action(self):
        """
        :py:meth:`elasticsearch.client.SnapshotClient.create` a snapshot of
        :py:attr:`indices`, with passed parameters.
        """
        if not self.skip_repo_fs_check:
            verify_repository(self.client, self.repository)
        if snapshot_running(self.client):
            raise SnapshotInProgress('Snapshot already in progress.')
        try:
            self.loggit.info(
                'Creating snapshot "%s" from indices: %s',
                self.name,
                self.index_list.indices,
            )
            # Always set wait_for_completion to False. Let 'wait_for_it' do its
            # thing if wait_for_completion is set to True. Report the task_id
            # either way.
            self.client.snapshot.create(
                repository=self.repository,
                snapshot=self.name,
                ignore_unavailable=self.ignore_unavailable,
                include_global_state=self.include_global_state,
                indices=self.indices,
                partial=self.partial,
                wait_for_completion=False,
            )
            if self.wait_for_completion:
                wait_for_it(
                    self.client,
                    'snapshot',
                    snapshot=self.name,
                    repository=self.repository,
                    wait_interval=self.wait_interval,
                    max_wait=self.max_wait,
                )
                self.report_state()
            else:
                msg = (
                    f'"wait_for_completion" set to {self.wait_for_completion}. '
                    f'Remember to check for successful completion manually.'
                )
                self.loggit.warning(msg)
        except Exception as err:
            report_failure(err)


class DeleteSnapshots:
    """Delete Snapshots Action Class"""

    def __init__(self, slo, retry_interval=120, retry_count=3):
        """
        :param slo: A SnapshotList object
        :type slo: :py:class:`~.curator.snapshotlist.SnapshotList`
        :param retry_interval: Seconds to delay betwen retries. (Default: ``120``)
        :type retry_interval: int
        :param retry_count: Number of attempts to make. (Default: ``3``)
        :type retry_count: int
        """
        verify_snapshot_list(slo)
        #: The :py:class:`~.curator.snapshotlist.SnapshotList` object passed from param
        #: ``slo``
        self.snapshot_list = slo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`snapshot_list`
        self.client = slo.client
        #: Object attribute that gets the value of param ``retry_interval``.
        self.retry_interval = retry_interval
        #: Object attribute that gets the value of param ``retry_count``.
        self.retry_count = retry_count
        #: Object attribute that gets its value from :py:attr:`snapshot_list`.
        self.repository = slo.repository
        self.loggit = logging.getLogger('curator.actions.delete_snapshots')

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        mykwargs = {
            'repository': self.repository,
            'retry_interval': self.retry_interval,
            'retry_count': self.retry_count,
        }
        for snap in self.snapshot_list.snapshots:
            self.loggit.info(
                'DRY-RUN: delete_snapshot: %s with arguments: %s', snap, mykwargs
            )

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.SnapshotClient.delete` snapshots in
        :py:attr:`snapshot_list`. Retry up to :py:attr:`retry_count` times, pausing
        :py:attr:`retry_interval` seconds between retries.
        """
        self.snapshot_list.empty_list_check()
        msg = (
            f'Deleting {len(self.snapshot_list.snapshots)} '
            f'selected snapshots: {self.snapshot_list.snapshots}'
        )
        self.loggit.info(msg)
        try:
            for snap in self.snapshot_list.snapshots:
                self.loggit.info('Deleting snapshot %s...', snap)
                self.client.snapshot.delete(repository=self.repository, snapshot=snap)
        # pylint: disable=broad-except
        except Exception as err:
            report_failure(err)


class Restore(object):
    """Restore Action Class

    Read more about identically named settings at:
    :py:meth:`elasticsearch.client.SnapshotClient.restore`
    """

    def __init__(
        self,
        slo,
        name=None,
        indices=None,
        include_aliases=False,
        ignore_unavailable=False,
        include_global_state=False,
        partial=False,
        rename_pattern=None,
        rename_replacement=None,
        extra_settings=None,
        wait_for_completion=True,
        wait_interval=9,
        max_wait=-1,
        skip_repo_fs_check=True,
    ):
        """
        :param slo: A SnapshotList object
        :param name: Name of the snapshot to restore.  If ``None``, use the most
            recent snapshot.
        :param indices: Indices to restore.  If ``None``, all in the snapshot
            will be restored.
        :param include_aliases: Restore aliases with the indices.
        :param ignore_unavailable: Ignore unavailable shards/indices.
        :param include_global_state: Restore cluster global state with snapshot.
        :param partial: Do not fail if primary shard is unavailable.
        :param rename_pattern: A regular expression pattern with one or more captures,
            e.g. ``index_(.+)``
        :param rename_replacement: A target index name pattern with `$#` numbered
            references to the captures in ``rename_pattern``, e.g.
            ``restored_index_$1``
        :param extra_settings: Index settings to apply to restored indices.
        :param wait_for_completion: Wait for completion before returning.
        :param wait_interval: Seconds to wait between completion checks.
        :param max_wait: Maximum number of seconds to ``wait_for_completion``
        :param skip_repo_fs_check: Do not validate write access to repository on
            all cluster nodes before proceeding. Useful for shared filesystems
            where intermittent timeouts can affect validation, but won't likely
            affect snapshot success. (Default: ``True``)

        :type slo: :py:class:`~.curator.snapshotlist.SnapshotList`
        :type name: str
        :type indices: list
        :type include_aliases: bool
        :type ignore_unavailable: bool
        :type include_global_state: bool
        :type partial: bool
        :type rename_pattern: str
        :type rename_replacement: str
        :type extra_settings: dict
        :type wait_for_completion: bool
        :type wait_interval: int
        :type max_wait: int
        :type skip_repo_fs_check: bool
        """
        if extra_settings is None:
            extra_settings = {}
        self.loggit = logging.getLogger('curator.actions.snapshot')
        verify_snapshot_list(slo)
        # Get the most recent snapshot.
        most_recent = slo.most_recent()
        self.loggit.debug('"most_recent" snapshot: %s', most_recent)
        #: Object attribute that gets the value of param ``name`` if not ``None``,
        #: or the output from :py:meth:`~.curator.SnapshotList.most_recent`.
        self.name = name if name else most_recent
        # Stop here now, if it's not a successful snapshot.
        if slo.snapshot_info[self.name]['state'] == 'PARTIAL' and partial:
            self.loggit.warning('Performing restore of snapshot in state PARTIAL.')
        elif slo.snapshot_info[self.name]['state'] != 'SUCCESS':
            raise CuratorException(
                'Restore operation can only be performed on snapshots with '
                'state "SUCCESS", or "PARTIAL" if partial=True.'
            )

        #: Internal reference to `slo`
        self.snapshot_list = slo
        #: The :py:class:`~.elasticsearch.Elasticsearch` client object derived from
        #: :py:attr:`snapshot_list`
        self.client = slo.client
        #: Object attribute that gets the value of ``repository`` from
        #: :py:attr:`snapshot_list`.
        self.repository = slo.repository
        self.loggit.debug('indices: %s', indices)
        if indices:
            self.indices = ensure_list(indices)
        else:
            self.indices = slo.snapshot_info[self.name]['indices']
        self.loggit.debug('self.indices: %s', self.indices)
        #: Object attribute that gets the value of param ``wait_for_completion``.
        self.wfc = wait_for_completion
        #: Object attribute that gets the value of param ``wait_interval``.
        self.wait_interval = wait_interval
        #: Object attribute that gets the value of param ``max_wait``.
        self.max_wait = max_wait
        #: Object attribute that gets the value of param ``rename_pattern``.
        #: Empty :py:class:`str` if ``None``
        self.rename_pattern = rename_pattern if rename_replacement is not None else ''
        #: Object attribute that gets the value of param ``rename_replacement``. Empty
        #: :py:class:`str` if ``None``
        self.rename_replacement = (
            rename_replacement if rename_replacement is not None else ''
        )
        #: Object attribute derived from :py:attr:`rename_replacement`. but with Java
        #: regex group designations of ``$#`` converted to Python's ``\\#`` style.
        self.py_rename_replacement = self.rename_replacement.replace('$', '\\')
        #: Object attribute that gets the value of param ``max_wait``.
        self.skip_repo_fs_check = skip_repo_fs_check

        #: Object attribute that gets populated from other params/attributes.
        #: Deprecated, but not removed. Lazy way to keep from updating
        #: :py:meth:`do_dry_run`. Will fix later.
        self.body = {
            'indices': self.indices,
            'include_aliases': include_aliases,
            'ignore_unavailable': ignore_unavailable,
            'include_global_state': include_global_state,
            'partial': partial,
            'rename_pattern': self.rename_pattern,
            'rename_replacement': self.rename_replacement,
        }
        #: Object attribute that gets the value of param ``include_aliases``.
        self.include_aliases = include_aliases
        #: Object attribute that gets the value of param ``ignore_unavailable``.
        self.ignore_unavailable = ignore_unavailable
        #: Object attribute that gets the value of param ``include_global_state``.
        self.include_global_state = include_global_state
        #: Object attribute that gets the value of param ``include_aliases``.
        self.include_aliases = include_aliases
        #: Object attribute that gets the value of param ``partial``.
        self.partial = partial
        #: Object attribute that gets the value of param ``extra_settings``.
        self.index_settings = None

        if extra_settings:
            self.loggit.debug(
                'Adding extra_settings to restore body: %s', extra_settings
            )
            self.index_settings = extra_settings
            try:
                self.body.update(extra_settings)
            except Exception:
                self.loggit.error('Unable to apply extra settings to restore body')
        self.loggit.debug('REPOSITORY: %s', self.repository)
        self.loggit.debug('WAIT_FOR_COMPLETION: %s', self.wfc)
        self.loggit.debug('SKIP_REPO_FS_CHECK: %s', self.skip_repo_fs_check)
        self.loggit.debug('BODY: %s', self.body)
        self._get_expected_output()

    def _get_expected_output(self):
        if self.indices == self.snapshot_list.snapshot_info[self.name]['indices']:
            indices = self.indices
        else:
            indices = multitarget_match(
                to_csv(self.indices),
                self.snapshot_list.snapshot_info[self.name]['indices'],
            )
        if not self.rename_pattern and not self.rename_replacement:
            self.expected_output = indices
            self.loggit.debug('Expected output: %s', indices)
            return  # Don't stick around if we're not replacing anything
        self.expected_output = []
        for index in indices:
            self.expected_output.append(
                re.sub(self.rename_pattern, self.py_rename_replacement, index)
            )
            msg = f'index: {index} replacement: {self.expected_output[-1]}'
            self.loggit.debug(msg)

    def report_state(self):
        """
        Log the state of the restore. This should only be done if
        ``wait_for_completion`` is ``True``, and only after completing the restore.
        """
        all_indices = get_indices(self.client)
        self.loggit.debug('All indices: %s', all_indices)
        self.loggit.debug('Expected output: %s', self.expected_output)
        found_count = 0
        missing = []
        for index in self.expected_output:
            if index in all_indices:
                found_count += 1
                self.loggit.info('Found restored index %s', index)
            else:
                missing.append(index)
        if found_count == len(self.expected_output):
            self.loggit.info('All indices appear to have been restored.')
        else:
            msg = (
                f'Some of the indices do not appear to have been restored. '
                f'Missing: {missing}'
            )
            self.loggit.error(msg)
            raise FailedRestore(msg)

    def do_dry_run(self):
        """Log what the output would be, but take no action."""
        self.loggit.info('DRY-RUN MODE.  No changes will be made.')
        args = {'wait_for_completion': self.wfc, 'body': self.body}
        msg = (
            f'DRY-RUN: restore: Repository: {self.repository} '
            f'Snapshot name: {self.name} Arguments: {args}'
        )
        self.loggit.info(msg)

        for index in self.indices:
            if self.rename_pattern and self.rename_replacement:
                _ = re.sub(self.rename_pattern, self.py_rename_replacement, index)
                rmsg = f'as {_}'
            else:
                rmsg = ''
            self.loggit.info('DRY-RUN: restore: Index %s %s', index, rmsg)

    def do_action(self):
        """
        :py:meth:`~.elasticsearch.client.SnapshotClient.restore` :py:attr:`indices` from
        :py:attr:`name` with passed params.
        """
        if not self.skip_repo_fs_check:
            verify_repository(self.client, self.repository)
        if snapshot_running(self.client):
            raise SnapshotInProgress('Cannot restore while a snapshot is in progress.')
        try:
            self.loggit.info(
                'Restoring indices "%s" from snapshot: %s', self.indices, self.name
            )
            # Always set wait_for_completion to False. Let 'wait_for_it' do its
            # thing if wait_for_completion is set to True. Report the task_id
            # either way.
            self.client.snapshot.restore(
                repository=self.repository,
                snapshot=self.name,
                ignore_index_settings=None,
                ignore_unavailable=self.ignore_unavailable,
                include_aliases=self.include_aliases,
                include_global_state=self.include_global_state,
                index_settings=self.index_settings,
                indices=self.indices,
                partial=self.partial,
                rename_pattern=self.rename_pattern,
                rename_replacement=self.rename_replacement,
                wait_for_completion=False,
            )
            if self.wfc:
                wait_for_it(
                    self.client,
                    'restore',
                    index_list=self.expected_output,
                    wait_interval=self.wait_interval,
                    max_wait=self.max_wait,
                )
                self.report_state()
            else:
                msg = (
                    f'"wait_for_completion" set to {self.wfc}. '
                    f'Remember to check for successful completion manually.'
                )
                self.loggit.warning(msg)
        except Exception as err:
            report_failure(err)



"""Index Rollover Singleton"""

import click
from es_client.utils import prune_nones
from curator.cli_singletons.object_class import CLIAction
from curator.cli_singletons.utils import json_to_dict


# pylint: disable=line-too-long
@click.command()
@click.option('--name', type=str, help='Alias name', required=True)
@click.option('--max_age', type=str, help='max_age condition value (see documentation)')
@click.option(
    '--max_docs', type=str, help='max_docs condition value (see documentation)'
)
@click.option(
    '--max_size', type=str, help='max_size condition value (see documentation)'
)
@click.option(
    '--extra_settings',
    type=str,
    help='JSON version of extra_settings (see documentation)',
    callback=json_to_dict,
)
@click.option(
    '--new_index', type=str, help='Optional new index name (see documentation)'
)
@click.option(
    '--wait_for_active_shards',
    type=int,
    default=1,
    show_default=True,
    help='Wait for number of shards to be active before returning',
)
@click.option(
    '--allow_ilm_indices/--no-allow_ilm_indices',
    help='Allow Curator to operate on Index Lifecycle Management monitored indices.',
    default=False,
    show_default=True,
)
@click.option(
    '--include_hidden/--no-include_hidden',
    help='Allow Curator to operate on hidden indices (and data_streams).',
    default=False,
    show_default=True,
)
@click.pass_context
def rollover(
    ctx,
    name,
    max_age,
    max_docs,
    max_size,
    extra_settings,
    new_index,
    wait_for_active_shards,
    allow_ilm_indices,
    include_hidden,
):
    """
    Rollover Index associated with Alias
    """
    conditions = prune_nones(
        {
            'max_age': max_age,
            'max_docs': max_docs,
            'max_size': max_size,
        }
    )
    manual_options = {
        'name': name,
        'conditions': conditions,
        'allow_ilm_indices': allow_ilm_indices,
        'include_hidden': include_hidden,
    }
    # ctx.info_name is the name of the function or name specified in
    # @click.command decorator
    action = CLIAction(
        ctx.info_name,
        ctx.obj['configdict'],
        manual_options,
        [],
        True,
        extra_settings=extra_settings,
        new_index=new_index,
        wait_for_active_shards=wait_for_active_shards,
    )
    action.do_singleton_action(dry_run=ctx.obj['dry_run'])



"""Test reindex action functionality"""

# pylint: disable=C0115, C0116, invalid-name
import os
from unittest.case import SkipTest
import pytest
from dotmap import DotMap  # type: ignore
from es_client.builder import Builder
from curator.helpers.getters import get_indices
from . import CuratorTestCase
from . import testvars

UNDEF = 'undefined'
HOST = os.environ.get('TEST_ES_SERVER', 'http://127.0.0.1:9200')
RHOST = os.environ.get('REMOTE_ES_SERVER', UNDEF)
WAIT_INTERVAL = 1
MAX_WAIT = 3


class TestActionFileReindex(CuratorTestCase):
    """Test file-based reindex operations"""

    def test_reindex_manual(self):
        """Test that manual reindex results in proper count of documents"""
        source = 'my_source'
        dest = 'my_dest'
        expected = 3
        self.create_index(source)
        self.add_docs(source)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex.format(WAIT_INTERVAL, MAX_WAIT, source, dest),
        )
        self.invoke_runner()
        assert expected == self.client.count(index=dest)['count']

    def test_reindex_selected(self):
        """Reindex selected indices"""
        source = 'my_source'
        dest = 'my_dest'
        expected = 3
        self.create_index(source)
        self.add_docs(source)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex.format(WAIT_INTERVAL, MAX_WAIT, 'REINDEX_SELECTION', dest),
        )
        self.invoke_runner()
        assert expected == self.client.count(index=dest)['count']

    def test_reindex_empty_list(self):
        """
        This test raises <class 'curator.exceptions.FailedExecution'>:
        Reindex failed. The index or alias identified by "my_dest" was not found.
        The source is never created, so the dest is also not able to be created
        This means that we expect an empty list.
        """
        source = 'my_source'
        dest = 'my_dest'
        expected = []
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex.format(WAIT_INTERVAL, MAX_WAIT, source, dest),
        )
        self.invoke_runner()
        assert expected == get_indices(self.client)

    def test_reindex_selected_many_to_one(self):
        """Test reindexing many indices to one destination"""
        source1 = 'my_source1'
        source2 = 'my_source2'
        dest = 'my_dest'
        expected = 6
        self.create_index(source1)
        self.add_docs(source1)
        self.create_index(source2)
        for i in ["4", "5", "6"]:
            self.client.create(
                index=source2, id=i, document={"doc" + i: 'TEST DOCUMENT'}
            )
            # Decorators make this pylint exception necessary
            # pylint: disable=E1123
            self.client.indices.flush(index=source2, force=True)
            self.client.indices.refresh(index=source2)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex.format(WAIT_INTERVAL, MAX_WAIT, 'REINDEX_SELECTION', dest),
        )
        self.invoke_runner()
        self.client.indices.refresh(index=dest)
        assert expected == self.client.count(index=dest)['count']

    def test_reindex_selected_empty_list_fail(self):
        """Ensure an empty list results in an exit code 1"""
        source1 = 'my_source1'
        source2 = 'my_source2'
        dest = 'my_dest'
        self.create_index(source1)
        self.add_docs(source1)
        self.create_index(source2)
        for i in ["4", "5", "6"]:
            self.client.create(
                index=source2, id=i, document={"doc" + i: 'TEST DOCUMENT'}
            )
            # Decorators make this pylint exception necessary
            # pylint: disable=E1123
            self.client.indices.flush(index=source2, force=True)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex_empty_list.format('false', WAIT_INTERVAL, MAX_WAIT, dest),
        )
        self.invoke_runner()
        assert 1 == self.result.exit_code

    def test_reindex_selected_empty_list_pass(self):
        """Ensure an empty list results in an exit code 0"""
        source1 = 'my_source1'
        source2 = 'my_source2'
        dest = 'my_dest'
        self.create_index(source1)
        self.add_docs(source1)
        self.create_index(source2)
        for i in ["4", "5", "6"]:
            self.client.create(
                index=source2, id=i, document={"doc" + i: 'TEST DOCUMENT'}
            )
            # Decorators make this pylint exception necessary
            # pylint: disable=E1123
            self.client.indices.flush(index=source2, force=True)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex_empty_list.format('true', WAIT_INTERVAL, MAX_WAIT, dest),
        )
        self.invoke_runner()
        assert 0 == self.result.exit_code

    @pytest.mark.skipif((RHOST == UNDEF), reason='REMOTE_ES_SERVER is not defined')
    def test_reindex_from_remote(self):
        """Test remote reindex functionality"""
        diff_wait = 6
        source1 = 'my_source1'
        source2 = 'my_source2'
        prefix = 'my_'
        dest = 'my_dest'
        expected = 6
        # Build remote client
        try:
            remote_args = DotMap({'hosts': RHOST})
            remote_config = {'elasticsearch': {'client': remote_args.toDict()}}
            builder = Builder(configdict=remote_config)
            builder.version_min = (5, 0, 0)
            builder.connect()
            rclient = builder.client
            rclient.info()
        except Exception as exc:
            raise SkipTest(f'Unable to connect to host at {RHOST}: {exc}') from exc
        # Build indices remotely.
        counter = 0
        rclient.indices.delete(index=f'{source1},{source2}', ignore_unavailable=True)
        for rindex in [source1, source2]:
            rclient.indices.create(index=rindex)
            for i in range(0, 3):
                rclient.create(
                    index=rindex,
                    id=str(counter + 1),
                    document={"doc" + str(i): 'TEST DOCUMENT'},
                )
                counter += 1
                # Decorators make this pylint exception necessary
                # pylint: disable=E1123
                rclient.indices.flush(index=rindex, force=True)
                rclient.indices.refresh(index=rindex)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.remote_reindex.format(
                WAIT_INTERVAL, diff_wait, RHOST, 'REINDEX_SELECTION', dest, prefix
            ),
        )
        self.invoke_runner()
        # Do our own cleanup here.
        rclient.indices.delete(index=f'{source1},{source2}', ignore_unavailable=True)
        assert expected == self.client.count(index=dest)['count']

    @pytest.mark.skipif((RHOST == UNDEF), reason='REMOTE_ES_SERVER is not defined')
    def test_reindex_migrate_from_remote(self):
        """Test remote reindex migration"""
        source1 = 'my_source1'
        source2 = 'my_source2'
        prefix = 'my_'
        dest = 'MIGRATION'
        expected = 3
        # Build remote client
        try:
            remote_args = DotMap({'hosts': RHOST})
            remote_config = {'elasticsearch': {'client': remote_args.toDict()}}
            builder = Builder(configdict=remote_config)
            builder.version_min = (5, 0, 0)
            builder.connect()
            rclient = builder.client
            rclient.info()
        except Exception as exc:
            raise SkipTest(f'Unable to connect to host at {RHOST}: {exc}') from exc
        # Build indices remotely.
        counter = 0
        rclient.indices.delete(index=f'{source1},{source2}', ignore_unavailable=True)
        for rindex in [source1, source2]:
            rclient.indices.create(index=rindex)
            for i in range(0, 3):
                rclient.create(
                    index=rindex,
                    id=str(counter + 1),
                    document={"doc" + str(i): 'TEST DOCUMENT'},
                )
                counter += 1
                # Decorators make this pylint exception necessary
                # pylint: disable=E1123
                rclient.indices.flush(index=rindex, force=True)
                rclient.indices.refresh(index=rindex)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.remote_reindex.format(
                WAIT_INTERVAL, MAX_WAIT, RHOST, 'REINDEX_SELECTION', dest, prefix
            ),
        )
        self.invoke_runner()
        # Do our own cleanup here.
        rclient.indices.delete(index=f'{source1},{source2}', ignore_unavailable=True)
        # And now the neat trick of verifying that the reindex worked to both
        # indices, and they preserved their names
        assert expected == self.client.count(index=source1)['count']
        assert expected == self.client.count(index=source2)['count']

    @pytest.mark.skipif((RHOST == UNDEF), reason='REMOTE_ES_SERVER is not defined')
    def test_reindex_migrate_from_remote_with_pre_suf_fixes(self):
        """Ensure migrate from remote with prefixes and suffixes works"""
        source1 = 'my_source1'
        source2 = 'my_source2'
        prefix = 'my_'
        dest = 'MIGRATION'
        expected = 3
        mpfx = 'pre-'
        msfx = '-fix'
        # Build remote client
        try:
            remote_args = DotMap({'hosts': RHOST})
            remote_config = {'elasticsearch': {'client': remote_args.toDict()}}
            builder = Builder(configdict=remote_config)
            builder.version_min = (5, 0, 0)
            builder.connect()
            rclient = builder.client
            rclient.info()
        except Exception as exc:
            raise SkipTest(f'Unable to connect to host at {RHOST}: {exc}') from exc
        # Build indices remotely.
        counter = 0
        rclient.indices.delete(index=f'{source1},{source2}', ignore_unavailable=True)
        for rindex in [source1, source2]:
            rclient.indices.create(index=rindex)
            for i in range(0, 3):
                rclient.create(
                    index=rindex,
                    id=str(counter + 1),
                    document={"doc" + str(i): 'TEST DOCUMENT'},
                )
                counter += 1
                # Decorators make this pylint exception necessary
                # pylint: disable=E1123
                rclient.indices.flush(index=rindex, force=True)
                rclient.indices.refresh(index=rindex)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.migration_reindex.format(
                WAIT_INTERVAL,
                MAX_WAIT,
                mpfx,
                msfx,
                RHOST,
                'REINDEX_SELECTION',
                dest,
                prefix,
            ),
        )
        self.invoke_runner()
        # Do our own cleanup here.
        rclient.indices.delete(index=f'{source1},{source2}')
        # And now the neat trick of verifying that the reindex worked to both
        # indices, and they preserved their names
        assert expected == self.client.count(index=f'{mpfx}{source1}{msfx}')['count']

    def test_reindex_from_remote_no_connection(self):
        """Ensure that the inability to connect to the remote cluster fails"""
        dest = 'my_dest'
        bad_remote = 'http://127.0.0.1:9601'
        expected = 1
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.remote_reindex.format(
                WAIT_INTERVAL, MAX_WAIT, bad_remote, 'REINDEX_SELECTION', dest, 'my_'
            ),
        )
        self.invoke_runner()
        assert expected == self.result.exit_code

    @pytest.mark.skipif((RHOST == UNDEF), reason='REMOTE_ES_SERVER is not defined')
    def test_reindex_from_remote_no_indices(self):
        """
        Test that attempting to reindex remotely with an empty list exits with a fail
        """
        source1 = 'wrong1'
        source2 = 'wrong2'
        prefix = 'my_'
        dest = 'my_dest'
        expected = 1
        # Build remote client
        try:
            remote_args = DotMap({'hosts': RHOST})
            remote_config = {'elasticsearch': {'client': remote_args.toDict()}}
            builder = Builder(configdict=remote_config)
            builder.version_min = (5, 0, 0)
            builder.connect()
            rclient = builder.client
            rclient.info()
        except Exception as exc:
            raise SkipTest(f'Unable to connect to host at {RHOST}: {exc}') from exc
        # Build indices remotely.
        counter = 0
        # Force remove my_source1 and my_source2 to prevent false positives
        rclient.indices.delete(
            index=f"{'my_source1'},{'my_source2'}", ignore_unavailable=True
        )
        rclient.indices.delete(index=f'{source1},{source2}', ignore_unavailable=True)
        for rindex in [source1, source2]:
            rclient.indices.create(index=rindex)
            for i in range(0, 3):
                rclient.create(
                    index=rindex,
                    id=str(counter + 1),
                    document={"doc" + str(i): 'TEST DOCUMENT'},
                )
                counter += 1
                rclient.indices.flush(index=rindex, force=True)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.remote_reindex.format(
                WAIT_INTERVAL, MAX_WAIT, f'{RHOST}', 'REINDEX_SELECTION', dest, prefix
            ),
        )
        self.invoke_runner()
        # Do our own cleanup here.
        rclient.indices.delete(index=f'{source1},{source2}')
        assert expected == self.result.exit_code

    def test_reindex_into_alias(self):
        """Ensure that reindexing into an alias works as expected"""
        source = 'my_source'
        dest = 'my_dest'
        expected = 3
        alias_dict = {dest: {}}
        self.client.indices.create(index='dummy', aliases=alias_dict)
        self.add_docs(source)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex.format(WAIT_INTERVAL, MAX_WAIT, source, dest),
        )
        self.invoke_runner()
        assert expected == self.client.count(index=dest)['count']

    def test_reindex_manual_date_math(self):
        """Ensure date math is functional with reindex calls"""
        source = '<source-{now/d}>'
        dest = '<target-{now/d}>'
        expected = 3

        self.create_index(source)
        self.add_docs(source)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex.format(WAIT_INTERVAL, MAX_WAIT, source, dest),
        )
        self.invoke_runner()
        assert expected == self.client.count(index=dest)['count']

    def test_reindex_bad_mapping(self):
        """This test addresses GitHub issue #1260"""
        source = 'my_source'
        dest = 'my_dest'
        expected = 1
        settings = {"number_of_shards": 1, "number_of_replicas": 0}
        mappings1 = {"properties": {"doc1": {"type": "keyword"}}}
        mappings2 = {"properties": {"doc1": {"type": "integer"}}}
        self.client.indices.create(index=source, settings=settings, mappings=mappings1)
        self.add_docs(source)
        # Create the dest index with a different mapping.
        self.client.indices.create(index=dest, settings=settings, mappings=mappings2)
        self.write_config(self.args['configfile'], testvars.client_config.format(HOST))
        self.write_config(
            self.args['actionfile'],
            testvars.reindex.format(WAIT_INTERVAL, MAX_WAIT, source, dest),
        )
        self.invoke_runner()
        assert expected == self.result.exit_code



"""Singleton Utils Module"""

import json
from click import BadParameter
from es_client.utils import ensure_list


# Click functions require ctx and param to be passed positionally even if not used
# pylint: disable=unused-argument
def json_to_dict(ctx, param, value):
    """Convert JSON to dictionary"""
    if value is None:
        return {}
    try:
        return json.loads(value)
    except ValueError as exc:
        raise BadParameter(f'Invalid JSON for {param}: {value}') from exc


# Click functions require ctx and param to be passed positionally even if not used
# pylint: disable=unused-argument
def validate_filter_json(ctx, param, value):
    """Validate the JSON provided from the command-line"""
    # The "None" edge case should only be for optional filters, like alias add/remove
    if value is None:
        return value
    try:
        filter_list = ensure_list(json.loads(value))
        return filter_list
    except ValueError as exc:
        raise BadParameter(f'Filter list is invalid JSON: {value}') from exc



"""Curator date and time functions"""

import logging
import random
import re
import string
import time
from datetime import timedelta, datetime, timezone
from elasticsearch9.exceptions import NotFoundError
from curator.exceptions import ConfigurationError
from curator.defaults.settings import date_regex


class TimestringSearch:
    """
    An object to allow repetitive search against a string, ``searchme``, without
    having to repeatedly recreate the regex.

    :param timestring: An ``strftime`` pattern
    :type timestring: :py:func:`~.time.strftime`
    """

    def __init__(self, timestring):
        # pylint: disable=consider-using-f-string
        regex = r'(?P<date>{0})'.format(get_date_regex(timestring))

        #: Object attribute. ``re.compile(regex)`` where
        #: ``regex = r'(?P<date>{0})'.format(get_date_regex(timestring))``. Uses
        #: :py:func:`get_date_regex`
        self.pattern = re.compile(regex)
        #: Object attribute preserving param ``timestring``
        self.timestring = timestring

    def get_epoch(self, searchme):
        """
        :param searchme: A string to be matched against :py:attr:`pattern` that matches
            :py:attr:`timestring`

        :returns: The epoch timestamp extracted from ``searchme`` by regex matching
            against :py:attr:`pattern`
        :rtype: int or None
        """
        match = self.pattern.search(searchme)
        if match:
            if match.group("date"):
                timestamp = match.group("date")
                return datetime_to_epoch(get_datetime(timestamp, self.timestring))
            return None
        return None


def absolute_date_range(
    unit, date_from, date_to, date_from_format=None, date_to_format=None
):
    """
    This function calculates a date range with an absolute time stamp for both the
    start time and the end time. These dates are converted to epoch time. The parameter
    ``unit`` is used when the same simplified date is used for both ``date_from`` and
    ``date_to`` to calculate the duration. For example, if ``unit`` is ``months``, and
    ``date_from`` and ``date_to`` are both ``2017.01``, then the entire month of
    January 2017 will be the absolute date range.

    :param unit: One of ``hours``, ``days``, ``weeks``, ``months``, or ``years``.
    :param date_from: The simplified date for the start of the range
    :param date_to: The simplified date for the end of the range.
    :param date_from_format: The :py:func:`~.time.strftime` string used to parse
        ``date_from``
    :param date_to_format: The :py:func:`~.time.strftime` string used to parse
        ``date_to``

    :type unit: str
    :type date_from: str
    :type date_to: str
    :type date_from_format: str
    :type date_to_format: str

    :returns: The epoch start time and end time of a date range
    :rtype: tuple
    """
    logger = logging.getLogger(__name__)
    acceptable_units = [
        'seconds',
        'minutes',
        'hours',
        'days',
        'weeks',
        'months',
        'years',
    ]
    if unit not in acceptable_units:
        raise ConfigurationError(f'"unit" must be one of: {acceptable_units}')
    if not date_from_format or not date_to_format:
        raise ConfigurationError('Must provide "date_from_format" and "date_to_format"')
    try:
        start_epoch = datetime_to_epoch(get_datetime(date_from, date_from_format))
        logger.debug('Start ISO8601 = %s', epoch2iso(start_epoch))
    except Exception as err:
        raise ConfigurationError(
            f'Unable to parse "date_from" {date_from} and "date_from_format" '
            f'{date_from_format}. Error: {err}'
        ) from err
    try:
        end_date = get_datetime(date_to, date_to_format)
    except Exception as err:
        raise ConfigurationError(
            f'Unable to parse "date_to" {date_to} and "date_to_format" '
            f'{date_to_format}. Error: {err}'
        ) from err
    # We have to iterate to one more month, and then subtract a second to get
    # the last day of the correct month
    if unit == 'months':
        month = end_date.month
        year = end_date.year
        if month == 12:
            year += 1
            month = 1
        else:
            month += 1
        new_end_date = datetime(year, month, 1, 0, 0, 0)
        end_epoch = datetime_to_epoch(new_end_date) - 1
    # Similarly, with years, we need to get the last moment of the year
    elif unit == 'years':
        new_end_date = datetime(end_date.year + 1, 1, 1, 0, 0, 0)
        end_epoch = datetime_to_epoch(new_end_date) - 1
    # It's not months or years, which have inconsistent reckoning...
    else:
        # This lets us use an existing method to simply add 1 more unit's worth
        # of seconds to get hours, days, or weeks, as they don't change
        # We use -1 as point of reference normally subtracts from the epoch
        # and we need to add to it, so we'll make it subtract a negative value.
        # Then, as before, subtract 1 to get the end of the period
        end_epoch = (
            get_point_of_reference(unit, -1, epoch=datetime_to_epoch(end_date)) - 1
        )

    logger.debug('End ISO8601 = %s', epoch2iso(end_epoch))
    return (start_epoch, end_epoch)


def date_range(unit, range_from, range_to, epoch=None, week_starts_on='sunday'):
    """
    This function calculates a date range with a distinct epoch time stamp of both the
    start time and the end time in counts of ``unit`` relative to the time at
    execution, if ``epoch`` is not set.

    If ``unit`` is ``weeks``, you can also determine when a week begins using
    ``week_starts_on``, which can be either ``sunday`` or ``monday``.

    :param unit: One of ``hours``, ``days``, ``weeks``, ``months``, or ``years``.
    :param range_from: Count of ``unit`` in the past/future is the origin?
    :param range_to: Count of ``unit`` in the past/future is the end point?
    :param epoch: An epoch timestamp used to establish a point of reference for
        calculations.
    :param week_starts_on: Either ``sunday`` or ``monday``. Default is ``sunday``

    :type unit: str
    :type range_from: int
    :type range_to: int
    :type epoch: int
    :type week_starts_on: str

    :returns: The epoch start time and end time of a date range
    :rtype: tuple
    """
    logger = logging.getLogger(__name__)
    start_date = None
    start_delta = None
    acceptable_units = ['hours', 'days', 'weeks', 'months', 'years']
    if unit not in acceptable_units:
        raise ConfigurationError(f'"unit" must be one of: {acceptable_units}')
    if not range_to >= range_from:
        raise ConfigurationError(
            '"range_to" must be greater than or equal to "range_from"'
        )
    if not epoch:
        epoch = time.time()
    epoch = fix_epoch(epoch)
    raw_point_of_ref = datetime.fromtimestamp(epoch, timezone.utc)
    logger.debug('Raw point of Reference = %s', raw_point_of_ref)
    # Reverse the polarity, because -1 as last week makes sense when read by
    # humans, but datetime timedelta math makes -1 in the future.
    origin = range_from * -1
    # These if statements help get the start date or start_delta
    if unit == 'hours':
        point_of_ref = datetime(
            raw_point_of_ref.year,
            raw_point_of_ref.month,
            raw_point_of_ref.day,
            raw_point_of_ref.hour,
            0,
            0,
        )
        start_delta = timedelta(hours=origin)
    if unit == 'days':
        point_of_ref = datetime(
            raw_point_of_ref.year, raw_point_of_ref.month, raw_point_of_ref.day, 0, 0, 0
        )
        start_delta = timedelta(days=origin)
    if unit == 'weeks':
        point_of_ref = datetime(
            raw_point_of_ref.year, raw_point_of_ref.month, raw_point_of_ref.day, 0, 0, 0
        )
        sunday = False
        if week_starts_on.lower() == 'sunday':
            sunday = True
        weekday = point_of_ref.weekday()
        # Compensate for ISO week starting on Monday by default
        if sunday:
            weekday += 1
        logger.debug('Weekday = %s', weekday)
        start_delta = timedelta(days=weekday, weeks=origin)
    if unit == 'months':
        point_of_ref = datetime(
            raw_point_of_ref.year, raw_point_of_ref.month, 1, 0, 0, 0
        )
        year = raw_point_of_ref.year
        month = raw_point_of_ref.month
        if origin > 0:
            for _ in range(0, origin):
                if month == 1:
                    year -= 1
                    month = 12
                else:
                    month -= 1
        else:
            for _ in range(origin, 0):
                if month == 12:
                    year += 1
                    month = 1
                else:
                    month += 1
        start_date = datetime(year, month, 1, 0, 0, 0)
    if unit == 'years':
        point_of_ref = datetime(raw_point_of_ref.year, 1, 1, 0, 0, 0)
        start_date = datetime(raw_point_of_ref.year - origin, 1, 1, 0, 0, 0)
    if unit not in ['months', 'years']:
        start_date = point_of_ref - start_delta
    # By this point, we know our start date and can convert it to epoch time
    start_epoch = datetime_to_epoch(start_date)
    logger.debug('Start ISO8601 = %s', epoch2iso(start_epoch))
    # This is the number of units we need to consider.
    count = (range_to - range_from) + 1
    # We have to iterate to one more month, and then subtract a second to get
    # the last day of the correct month
    if unit == 'months':
        month = start_date.month
        year = start_date.year
        for _ in range(0, count):
            if month == 12:
                year += 1
                month = 1
            else:
                month += 1
        end_date = datetime(year, month, 1, 0, 0, 0)
        end_epoch = datetime_to_epoch(end_date) - 1
    # Similarly, with years, we need to get the last moment of the year
    elif unit == 'years':
        end_date = datetime((raw_point_of_ref.year - origin) + count, 1, 1, 0, 0, 0)
        end_epoch = datetime_to_epoch(end_date) - 1
    # It's not months or years, which have inconsistent reckoning...
    else:
        # This lets us use an existing method to simply add unit * count seconds
        # to get hours, days, or weeks, as they don't change
        end_epoch = get_point_of_reference(unit, count * -1, epoch=start_epoch) - 1
    logger.debug('End ISO8601 = %s', epoch2iso(end_epoch))
    return (start_epoch, end_epoch)


def datetime_to_epoch(mydate):
    """
    Converts datetime into epoch seconds

    :param mydate: A Python datetime
    :type mydate: :py:class:`~.datetime.datetime`

    :returns: An epoch timestamp based on ``mydate``
    :rtype: int
    """
    tdelta = mydate - datetime(1970, 1, 1)
    return tdelta.seconds + tdelta.days * 24 * 3600


def epoch2iso(epoch: int) -> str:
    """
    Return an ISO8601 value for epoch

    :param epoch: An epoch timestamp
    :type epoch: int

    :returns: An ISO8601 timestamp
    :rtype: str
    """
    # Because Python 3.12 now requires non-naive timezone declarations, we must change.
    #
    # ## Example:
    # ## epoch == 1491256800
    # ##
    # ## The old way:
    # ##datetime.utcfromtimestamp(epoch)
    # ##   datetime.datetime(2017, 4, 3, 22, 0).isoformat()
    # ##   Result: 2017-04-03T22:00:00
    # ##
    # ## The new way:
    # ##     datetime.fromtimestamp(epoch, timezone.utc)
    # ##     datetime.datetime(
    # ##         2017, 4, 3, 22, 0, tzinfo=datetime.timezone.utc).isoformat()
    # ##     Result: 2017-04-03T22:00:00+00:00
    # ##
    # ## End Example
    #
    # Note that the +00:00 is appended now where we affirmatively declare the UTC
    # timezone
    #
    # As a result, we will use this function to prune away the timezone if it is +00:00
    # and replace it with Z, which is shorter Zulu notation for UTC (which
    # Elasticsearch uses)
    #
    # We are MANUALLY, FORCEFULLY declaring timezone.utc, so it should ALWAYS be
    # +00:00, but could in theory sometime show up as a Z, so we test for that.

    parts = datetime.fromtimestamp(epoch, timezone.utc).isoformat().split('+')
    if len(parts) == 1:
        if parts[0][-1] == 'Z':
            return parts[0]  # Our ISO8601 already ends with a Z for Zulu/UTC time
        return f'{parts[0]}Z'  # It doesn't end with a Z so we put one there
    if parts[1] == '00:00':
        return f'{parts[0]}Z'  # It doesn't end with a Z so we put one there
    return f'{parts[0]}+{parts[1]}'  # Fallback publishes the +TZ, whatever that was


def fix_epoch(epoch):
    """
    Fix value of ``epoch`` to be the count since the epoch in seconds only, which
    should be 10 or fewer digits long.

    :param epoch: An epoch timestamp, in epoch + milliseconds, or microsecond, or even
        nanoseconds.
    :type epoch: int

    :returns: An epoch timestamp in seconds only, based on ``epoch``
    :rtype: int
    """
    try:
        # No decimals allowed
        epoch = int(epoch)
    except Exception as err:
        raise ValueError(
            f'Bad epoch value. Unable to convert {epoch} to int. {err}'
        ) from err

    # If we're still using this script past January, 2038, we have bigger
    # problems than my hacky math here...
    if len(str(epoch)) <= 10:
        # Epoch is fine, no changes
        pass
    elif len(str(epoch)) > 10 and len(str(epoch)) <= 13:
        epoch = int(epoch / 1000)
    else:
        orders_of_magnitude = len(str(epoch)) - 10
        powers_of_ten = 10**orders_of_magnitude
        epoch = int(epoch / powers_of_ten)
    return epoch


def get_date_regex(timestring):
    """
    :param timestring: An ``strftime`` pattern
    :type timestring: :py:func:`~.time.strftime`

    :returns: A regex string based on a provided :py:func:`~.time.strftime`
        ``timestring``.
    :rtype: str
    """
    logger = logging.getLogger(__name__)
    prev, regex = ('', '')
    logger.debug('Provided timestring = "%s"', timestring)
    for idx, char in enumerate(timestring):
        logger.debug('Current character: %s Array position: %s', char, idx)
        if char == '%':
            pass
        elif char in date_regex() and prev == '%':
            regex += r'\d{' + date_regex()[char] + '}'
        elif char in ['.', '-']:
            regex += "\\" + char
        else:
            regex += char
        prev = char
    logger.debug('regex = %s', regex)
    return regex


def get_datemath(client, datemath, random_element=None):
    """
    :param client: A client connection object
    :param datemath: An elasticsearch datemath string
    :param random_element: This allows external randomization of the name and is only
        useful for tests so that you can guarantee the output because you provided the
        random portion.

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type datemath: :py:class:`~.datemath.datemath`
    :type random_element: str

    :returns: the parsed index name from ``datemath``
    :rtype: str
    """
    logger = logging.getLogger(__name__)
    if random_element is None:
        random_prefix = 'curator_get_datemath_function_' + ''.join(
            random.choice(string.ascii_lowercase) for _ in range(32)
        )
    else:
        random_prefix = 'curator_get_datemath_function_' + random_element
    datemath_dummy = f'<{random_prefix}-{datemath}>'
    # We both want and expect a 404 here (NotFoundError), since we have
    # created a 32 character random string to definitely be an unknown
    # index name.
    logger.debug('Random datemath string for extraction: %s', datemath_dummy)
    faux_index = ''
    try:
        client.indices.get(index=datemath_dummy)
    except NotFoundError as err:
        # This is the magic.  Elasticsearch still gave us the formatted
        # index name in the error results.
        faux_index = err.body['error']['index']
    logger.debug('Response index name for extraction: %s', faux_index)
    # Now we strip the random index prefix back out again
    # pylint: disable=consider-using-f-string
    pattern = r'^{0}-(.*)$'.format(random_prefix)
    regex = re.compile(pattern)
    try:
        # And return only the now-parsed date string
        return regex.match(faux_index).group(1)
    except AttributeError as exc:
        raise ConfigurationError(
            f'The rendered index "{faux_index}" does not contain a valid date pattern '
            f'or has invalid index name characters.'
        ) from exc


def get_datetime(index_timestamp, timestring):
    """
    :param index_timestamp: The index timestamp
    :param timestring: An ``strftime`` pattern

    :type index_timestamp: str
    :type timestring: :py:func:`~.time.strftime`

    :returns: The datetime extracted from the index name, which is the index creation
        time.
    :rtype: :py:class:`~.datetime.datetime`
    """
    # Compensate for week of year by appending '%w' to the timestring
    # and '1' (Monday) to index_timestamp
    iso_week_number = False
    if '%W' in timestring or '%U' in timestring or '%V' in timestring:
        timestring += '%w'
        index_timestamp += '1'
        if '%V' in timestring and '%G' in timestring:
            iso_week_number = True
            # Fake as so we read Greg format instead. We will process it later
            timestring = timestring.replace("%G", "%Y").replace("%V", "%W")
    elif '%m' in timestring:
        if '%d' not in timestring:
            timestring += '%d'
            index_timestamp += '1'

    mydate = datetime.strptime(index_timestamp, timestring)

    # Handle ISO time string
    if iso_week_number:
        mydate = handle_iso_week_number(mydate, timestring, index_timestamp)

    return mydate


def get_point_of_reference(unit, count, epoch=None):
    """
    :param unit: One of ``seconds``, ``minutes``, ``hours``, ``days``, ``weeks``,
        ``months``, or ``years``.
    :param unit_count: The number of ``units``. ``unit_count`` * ``unit`` will be
        calculated out to the relative number of seconds.
    :param epoch: An epoch timestamp used in conjunction with ``unit`` and
        ``unit_count`` to establish a point of reference for calculations.

    :type unit: str
    :type unit_count: int
    :type epoch: int

    :returns: A point-of-reference timestamp in epoch + milliseconds by deriving from a
        ``unit`` and a ``count``, and an optional reference timestamp, ``epoch``
    :rtype: int
    """
    if unit == 'seconds':
        multiplier = 1
    elif unit == 'minutes':
        multiplier = 60
    elif unit == 'hours':
        multiplier = 3600
    elif unit == 'days':
        multiplier = 3600 * 24
    elif unit == 'weeks':
        multiplier = 3600 * 24 * 7
    elif unit == 'months':
        multiplier = 3600 * 24 * 30
    elif unit == 'years':
        multiplier = 3600 * 24 * 365
    else:
        raise ValueError(f'Invalid unit: {unit}.')
    # Use this moment as a reference point, if one is not provided.
    if not epoch:
        epoch = time.time()
    epoch = fix_epoch(epoch)
    return epoch - multiplier * count


def get_unit_count_from_name(index_name, pattern):
    """
    :param index_name: An index name
    :param pattern: A regular expression pattern

    :type index_name: str
    :type pattern: str

    :returns: The unit count, if a match is able to be found in the name
    :rtype: int
    """
    if pattern is None:
        return None
    match = pattern.search(index_name)
    if match:
        try:
            return int(match.group(1))
        # pylint: disable=broad-except
        except Exception:
            return None
    else:
        return None


def handle_iso_week_number(mydate, timestring, index_timestamp):
    """
    :param mydate: A Python datetime
    :param timestring: An ``strftime`` pattern
    :param index_timestamp: The index timestamp

    :type mydate: :py:class:`~.datetime.datetime`
    :type timestring: :py:func:`~.time.strftime`
    :type index_timestamp: str

    :returns: The date of the previous week based on ISO week number
    :rtype: :py:class:`~.datetime.datetime`
    """
    date_iso = mydate.isocalendar()
    # iso_week_str = "{Y:04d}{W:02d}".format(Y=date_iso[0], W=date_iso[1])
    iso_week_str = f'{date_iso[0]:04d}{date_iso[1]:02d}'
    greg_week_str = datetime.strftime(mydate, "%Y%W")

    # Edge case 1: ISO week number is bigger than Greg week number.
    # Ex: year 2014, all ISO week numbers were 1 more than in Greg.
    if (
        iso_week_str > greg_week_str
        or
        # Edge case 2: 2010-01-01 in ISO: 2009.W53, in Greg: 2010.W00
        # For Greg converting 2009.W53 gives 2010-01-04, converting back
        # to same timestring gives: 2010.W01.
        datetime.strftime(mydate, timestring) != index_timestamp
    ):

        # Remove one week in this case
        mydate = mydate - timedelta(days=7)
    return mydate


def isdatemath(data):
    """
    :param data: An expression to validate as being datemath or not
    :type data: str

    :returns: ``True`` if ``data`` is a valid datemath expression, else ``False``
    :rtype: bool
    """
    logger = logging.getLogger(__name__)
    initial_check = r'^(.).*(.)$'
    regex = re.compile(initial_check)
    opener = regex.match(data).group(1)
    closer = regex.match(data).group(2)
    logger.debug('opener =  %s, closer = %s', opener, closer)
    if (opener == '<' and closer != '>') or (opener != '<' and closer == '>'):
        raise ConfigurationError('Incomplete datemath encapsulation in "< >"')
    if opener != '<' and closer != '>':
        return False
    return True


def parse_date_pattern(name):
    """
    Scan and parse ``name`` for :py:func:`~.time.strftime` strings, replacing them with
    the associated value when found, but otherwise returning lowercase values, as
    uppercase snapshot names are not allowed. It will detect if the first character is
    a ``<``, which would indicate ``name`` is going to be using Elasticsearch date math
    syntax, and skip accordingly.

    The :py:func:`~.time.strftime` identifiers that Curator currently recognizes as
    acceptable include:

    * ``Y``: A 4 digit year
    * ``y``: A 2 digit year
    * ``m``: The 2 digit month
    * ``W``: The 2 digit week of the year
    * ``d``: The 2 digit day of the month
    * ``H``: The 2 digit hour of the day, in 24 hour notation
    * ``M``: The 2 digit minute of the hour
    * ``S``: The 2 digit number of second of the minute
    * ``j``: The 3 digit day of the year

    :param name: A name, which can contain :py:func:`~.time.strftime` strings
    :type name: str

    :returns: The parsed date pattern
    :rtype: str
    """
    logger = logging.getLogger(__name__)
    prev, rendered = ('', '')
    logger.debug('Provided index name: %s', name)
    for idx, char in enumerate(name):
        logger.debug('Current character in provided name: %s, position: %s', char, idx)
        if char == '<':
            logger.info('"%s" is probably using Elasticsearch date math.', name)
            rendered = name
            break
        if char == '%':
            pass
        elif char in date_regex() and prev == '%':
            rendered += str(datetime.now(timezone.utc).strftime(f'%{char}'))
        else:
            rendered += char
        logger.debug('Partially rendered name: %s', rendered)
        prev = char
    logger.debug('Fully rendered name: %s', rendered)
    return rendered


def parse_datemath(client, value):
    """
    Validate that ``value`` looks like proper datemath. If it passes this test, then
    try to ship it to Elasticsearch for real. It may yet fail this test, and if it
    does, it will raise a :py:exc:`~.curator.exceptions.ConfigurationError` exception.
    If it passes, return the fully parsed string.

    :param client: A client connection object
    :param value: A string to check for datemath

    :type client: :py:class:`~.elasticsearch.Elasticsearch`
    :type value: str

    :returns: A datemath indexname, fully rendered by Elasticsearch
    :rtype: str
    """
    logger = logging.getLogger(__name__)
    if not isdatemath(value):
        return value
    # if we didn't return here, we can continue, no 'else' needed.
    logger.debug('Properly encapsulated, proceeding to next evaluation...')
    # Our pattern has 4 capture groups.
    # 1. Everything after the initial '<' up to the first '{', which we call ``prefix``
    # 2. Everything between the outermost '{' and '}', which we call ``datemath``
    # 3. An optional inner '{' and '}' containing a date formatter and potentially a
    #    timezone. Not captured.
    # 4. Everything after the last '}' up to the closing '>'
    pattern = r'^<([^\{\}]*)?(\{.*(\{.*\})?\})([^\{\}]*)?>$'
    regex = re.compile(pattern)
    try:
        prefix = regex.match(value).group(1) or ''
        datemath = regex.match(value).group(2)
        # formatter = regex.match(value).group(3) or '' (not captured, but counted)
        suffix = regex.match(value).group(4) or ''
    except AttributeError as exc:
        raise ConfigurationError(
            f'Value "{value}" does not contain a valid datemath pattern.'
        ) from exc

    return f'{prefix}{get_datemath(client, datemath)}{suffix}'
